"""Base emmet model to add default metadata."""

from typing import Literal

from pydantic import BaseModel, Field
from pymatgen.core import __version__ as pmg_version

from emmet.core import __version__
from emmet.core.types.typing import NullableDateTimeType


class EmmetMeta(BaseModel):
    """Default emmet metadata."""

    emmet_version: str | None = Field(
        __version__, description="The version of emmet this document was built with."
    )
    pymatgen_version: str | None = Field(
        pmg_version, description="The version of pymatgen this document was built with."
    )

    run_id: str | None = Field(
        None, description="The run id associated with this data build."
    )

    batch_id: str | None = Field(
        None,
        description="Identifier corresponding to the origin of this document's blessed task.",
    )

    database_version: str | None = Field(
        None, description="The database version for the built data."
    )

    build_date: NullableDateTimeType = Field(  # type: ignore
        description="The build date for this document.",
    )

    license: Literal["BY-C", "BY-NC"] | None = Field(
        None, description="License for the data entry."
    )


class EmmetBaseModel(BaseModel):
    """Base Model for default emmet data."""

    builder_meta: EmmetMeta | None = Field(
        default_factory=EmmetMeta,  # type: ignore
        description="Builder metadata.",
    )


class CalcMeta(BaseModel):
    """Lightweight representation for storing jobflow-style job info."""

    identifier: str | None = Field(
        None, description="The identifier (UUID, ULID, etc.) of the calculation."
    )
    dir_name: str | None = Field(
        None, description="The file-systm-like path where a job was run."
    )
    name: str | None = Field(
        None, description="The job name or a convenient label for it."
    )
    parents: list[str] | None = Field(
        None, description="A list of all UUIDs which fed into this job."
    )



from typing import Annotated, TypeVar

from pydantic import BeforeValidator, WrapSerializer
from pymatgen.phonon.ir_spectra import IRDielectricTensor
from typing_extensions import TypedDict

from emmet.core.types.pymatgen_types.structure_adapter import TypedStructureDict

TypedIRDTensorDict = TypedDict(
    "TypedIRDTensorDict",
    {
        "@module": str,
        "@class": str,
        "oscillator_strength": list[float],
        "ph_freqs_gamma": list[float],
        "structure": TypedStructureDict,
        "epsilon_infinity": list[float],
    },
)

IRDielectricTensorTypeVar = TypeVar(
    "IRDielectricTensorTypeVar", IRDielectricTensor, TypedIRDTensorDict
)

IRDielectricTensorType = Annotated[
    IRDielectricTensorTypeVar,
    BeforeValidator(
        lambda x: IRDielectricTensor.from_dict(x) if isinstance(x, dict) else x
    ),
    WrapSerializer(lambda x, nxt, info: x.as_dict(), return_type=TypedIRDTensorDict),
]



import pytest
from pymatgen.core import Structure
from robocrys import __version__ as __robocrys_version__

from . import test_structures
from emmet.core import ARROW_COMPATIBLE
from emmet.core.robocrys import RobocrystallogapherDoc

if ARROW_COMPATIBLE:
    import pyarrow as pa

    from emmet.core.arrow import arrowize


@pytest.mark.parametrize("structure", test_structures.values())
def test_robocrys(structure: Structure):
    """Very simple test to make sure this actually works"""
    print(f"Should work : {structure.composition}")
    doc = RobocrystallogapherDoc.from_structure(
        structure=structure,
        material_id=33,
        deprecated=False,
        robocrys_version=__robocrys_version__,
    )
    assert doc is not None


@pytest.mark.skipif(
    not ARROW_COMPATIBLE, reason="pyarrow must be installed to run this test."
)
def test_arrow():
    structure = next(iter(test_structures.values()))
    doc = RobocrystallogapherDoc.from_structure(
        structure=structure,
        material_id=33,
        deprecated=False,
        robocrys_version=__robocrys_version__,
    )
    arrow_struct = pa.scalar(
        doc.model_dump(context={"format": "arrow"}),
        type=arrowize(RobocrystallogapherDoc),
    )
    test_arrow_doc = RobocrystallogapherDoc(
        **arrow_struct.as_py(maps_as_pydicts="strict")
    )

    assert doc.model_dump() == test_arrow_doc.model_dump()



import pytest
from monty.serialization import loadfn
from pymatgen.core import Structure

from emmet.core import ARROW_COMPATIBLE
from emmet.core.magnetism import MagnetismDoc
from emmet.core.utils import jsanitize

if ARROW_COMPATIBLE:
    import pyarrow as pa

    from emmet.core.arrow import arrowize


@pytest.fixture(scope="module")
def magnetism_mats(test_dir):
    return loadfn(test_dir / "magnetism/magnetism_mats_sample.json.gz")


def test_magnetism_doc(magnetism_mats):
    test_orderings = {"mp-1034331": "FM", "mp-753472": "NM"}

    for material in magnetism_mats:
        structure = Structure.from_dict(jsanitize(material["structure"]))
        total_magnetization = material["magnetism"]["total_magnetization"]
        material_id = material["task_id"]

        doc = MagnetismDoc.from_structure(
            structure=structure,
            material_id=material_id,
            total_magnetization=total_magnetization,
            deprecated=False,
        )
        assert doc is not None
        assert doc.total_magnetization == pytest.approx(abs(total_magnetization))

        if material_id in test_orderings:
            assert doc.ordering == test_orderings[material_id]


@pytest.mark.skipif(
    not ARROW_COMPATIBLE, reason="pyarrow must be installed to run this test."
)
def test_arrow(magnetism_mats):
    doc = MagnetismDoc.from_structure(
        structure=Structure.from_dict(jsanitize(magnetism_mats[0]["structure"])),
        material_id=magnetism_mats[0]["task_id"],
        total_magnetization=magnetism_mats[0]["magnetism"]["total_magnetization"],
        deprecated=False,
    )
    arrow_struct = pa.scalar(
        doc.model_dump(context={"format": "arrow"}), type=arrowize(MagnetismDoc)
    )
    test_arrow_doc = MagnetismDoc(**arrow_struct.as_py(maps_as_pydicts="strict"))

    assert doc.model_dump() == test_arrow_doc.model_dump()



"""Core definition of Structure and Molecule metadata."""

from __future__ import annotations

from typing import TYPE_CHECKING, TypeVar

from pydantic import BaseModel, Field
from pymatgen.core.composition import Composition
from pymatgen.core.structure import Molecule, Structure

from emmet.core.symmetry import PointGroupData, SymmetryData
from emmet.core.types.pymatgen_types.composition_adapter import CompositionType
from emmet.core.types.pymatgen_types.element_adapter import ElementType
from emmet.core.utils import get_graph_hash

if TYPE_CHECKING:
    from typing_extensions import Self

T = TypeVar("T", bound="StructureMetadata")
S = TypeVar("S", bound="MoleculeMetadata")

try:
    from openbabel import openbabel
except Exception:
    openbabel = None


class StructureMetadata(BaseModel):
    """Mix-in class for structure metadata."""

    # Structure metadata
    nsites: int | None = Field(
        None, description="Total number of sites in the structure."
    )
    elements: list[ElementType] | None = Field(
        None, description="List of elements in the material."
    )
    nelements: int | None = Field(None, description="Number of elements.")
    composition: CompositionType | None = Field(
        None, description="Full composition for the material."
    )
    composition_reduced: CompositionType | None = Field(
        None,
        title="Reduced Composition",
        description="Simplified representation of the composition.",
    )
    formula_pretty: str | None = Field(
        None,
        title="Pretty Formula",
        description="Cleaned representation of the formula.",
    )
    formula_anonymous: str | None = Field(
        None,
        title="Anonymous Formula",
        description="Anonymized representation of the formula.",
    )
    chemsys: str | None = Field(
        None,
        title="Chemical System",
        description="dash-delimited string of elements in the material.",
    )
    volume: float | None = Field(
        None,
        title="Volume",
        description="Total volume for this structure in Angstroms^3.",
    )

    density: float | None = Field(
        None, title="Density", description="Density in grams per cm^3."
    )

    density_atomic: float | None = Field(
        None,
        title="Packing Density",
        description="The atomic packing density in atoms per cm^3.",
    )

    symmetry: SymmetryData | None = Field(
        None, description="Symmetry data for this material."
    )

    @classmethod
    def from_composition(
        cls,
        composition: Composition,
        fields: list[str] | None = None,
        **kwargs,
    ) -> Self:
        fields = (
            [
                "elements",
                "nelements",
                "composition",
                "composition_reduced",
                "formula_pretty",
                "formula_anonymous",
                "chemsys",
            ]
            if fields is None
            else fields
        )
        composition = composition.remove_charges()

        elsyms = sorted({e.symbol for e in composition.elements})

        data = {
            "elements": elsyms,
            "nelements": len(elsyms),
            "composition": composition,
            "composition_reduced": composition.reduced_composition.remove_charges(),
            "formula_pretty": composition.reduced_formula,
            "formula_anonymous": composition.anonymized_formula,
            "chemsys": "-".join(elsyms),
        }

        return cls(**{k: v for k, v in data.items() if k in fields}, **kwargs)  # type: ignore[arg-type]

    @classmethod
    def from_structure(
        cls,
        meta_structure: Structure,
        fields: list[str] | None = None,
        **kwargs,
    ) -> Self:
        fields = (
            [
                "nsites",
                "elements",
                "nelements",
                "composition",
                "composition_reduced",
                "formula_pretty",
                "formula_anonymous",
                "chemsys",
                "volume",
                "density",
                "density_atomic",
                "symmetry",
            ]
            if fields is None
            else fields
        )
        comp = meta_structure.composition.remove_charges()
        elsyms = sorted({e.symbol for e in comp.elements})
        symmetry = SymmetryData.from_structure(meta_structure)

        data = {
            "nsites": meta_structure.num_sites,
            "elements": elsyms,
            "nelements": len(elsyms),
            "composition": comp,
            "composition_reduced": comp.reduced_composition,
            "formula_pretty": comp.reduced_formula,
            "formula_anonymous": comp.anonymized_formula,
            "chemsys": "-".join(elsyms),
            "volume": meta_structure.volume,
            "density": meta_structure.density,
            "density_atomic": meta_structure.volume / meta_structure.num_sites,
            "symmetry": symmetry,
        }
        kwargs.update({k: v for k, v in data.items() if k in fields})
        return cls(**kwargs)


class MoleculeMetadata(BaseModel):
    """Mix-in class for molecule metadata."""

    charge: int | None = Field(None, description="Charge of the molecule")
    spin_multiplicity: int | None = Field(
        None, description="Spin multiplicity of the molecule"
    )
    natoms: int | None = Field(
        None, description="Total number of atoms in the molecule"
    )
    elements: list[ElementType] | None = Field(
        None, description="List of elements in the molecule"
    )
    nelements: int | None = Field(None, title="Number of Elements")
    nelectrons: int | None = Field(
        None,
        title="Number of electrons",
        description="The total number of electrons for the molecule",
    )
    composition: CompositionType | None = Field(
        None, description="Full composition for the molecule"
    )
    composition_reduced: CompositionType | None = Field(
        None,
        title="Reduced Composition",
        description="Simplified representation of the composition",
    )
    formula_alphabetical: str | None = Field(
        None,
        title="Alphabetical Formula",
        description="Alphabetical molecular formula",
    )
    formula_pretty: str | None = Field(
        None,
        title="Pretty Formula",
        description="Cleaned representation of the formula.",
    )
    formula_anonymous: str | None = Field(
        None,
        title="Anonymous Formula",
        description="Anonymized representation of the formula",
    )
    chemsys: str | None = Field(
        None,
        title="Chemical System",
        description="dash-delimited string of elements in the molecule",
    )
    symmetry: PointGroupData | None = Field(
        None, description="Symmetry data for this molecule"
    )
    species_hash: str | None = Field(
        None,
        description="Weisfeiler Lehman (WL) graph hash using the atom species as the graph "
        "node attribute.",
    )
    coord_hash: str | None = Field(
        None,
        description="Weisfeiler Lehman (WL) graph hash using the atom coordinates as the graph "
        "node attribute.",
    )

    @classmethod
    def from_composition(
        cls,
        comp: Composition,
        fields: list[str] | None = None,
        **kwargs,
    ) -> Self:
        """
        Create a MoleculeMetadata model from a composition.

        Parameters
        ----------
        comp : .Composition
            A pymatgen composition.
        fields : list of str or None
            Composition fields to include.
        **kwargs
            Keyword arguments that are passed to the model constructor.

        Returns
        -------
        T
            A molecule metadata model.
        """
        fields = (
            [
                "elements",
                "nelements",
                "composition",
                "composition_reduced",
                "formula_alphabetical",
                "formula_pretty",
                "formula_anonymous",
                "chemsys",
            ]
            if fields is None
            else fields
        )
        elsyms = sorted({e.symbol for e in comp.elements})

        data = {
            "elements": elsyms,
            "nelements": len(elsyms),
            "composition": comp,
            "composition_reduced": comp.reduced_composition,
            "formula_alphabetical": comp.alphabetical_formula,
            "formula_pretty": comp.reduced_formula,
            "formula_anonymous": comp.anonymized_formula,
            "chemsys": "-".join(elsyms),
        }

        return cls(**{k: v for k, v in data.items() if k in fields}, **kwargs)  # type: ignore[arg-type]

    @classmethod
    def from_molecule(
        cls,
        meta_molecule: Molecule,
        fields: list[str] | None = None,
        **kwargs,
    ) -> Self:
        fields = (
            [
                "charge",
                "spin_multiplicity",
                "natoms",
                "elements",
                "nelements",
                "nelectrons",
                "composition",
                "composition_reduced",
                "formula_alphabetical",
                "formula_pretty",
                "formula_anonymous",
                "chemsys",
                "symmetry",
                "species_hash",
                "coord_hash",
            ]
            if fields is None
            else fields
        )
        comp = meta_molecule.composition
        elsyms = sorted({e.symbol for e in comp.elements})
        symmetry = PointGroupData.from_molecule(meta_molecule)

        data = {
            "charge": int(meta_molecule.charge),
            "spin_multiplicity": meta_molecule.spin_multiplicity,
            "natoms": len(meta_molecule),
            "elements": elsyms,
            "nelements": len(elsyms),
            "nelectrons": int(meta_molecule.nelectrons),
            "composition": comp,
            "composition_reduced": comp.reduced_composition,
            "formula_alphabetical": comp.alphabetical_formula,
            "formula_pretty": comp.reduced_formula,
            "formula_anonymous": comp.anonymized_formula,
            "chemsys": "-".join(elsyms),
            "symmetry": symmetry,
        }
        if openbabel:
            data["species_hash"] = get_graph_hash(meta_molecule, "specie")
            data["coord_hash"] = get_graph_hash(meta_molecule, "coords")

        return cls(**{k: v for k, v in data.items() if k in fields}, **kwargs)



from pathlib import Path
from monty.serialization import dumpfn

from emmet.core.types.enums import (
    DocEnum,
    ValueEnum,
)


def test_value_enum(tmp_path):
    class TempEnum(ValueEnum):
        A = "A"
        B = "B"

    assert str(TempEnum.A) == "A"
    assert str(TempEnum.B) == "B"

    dumpfn(TempEnum, tmp_path / "temp.json")
    assert Path(tmp_path, "temp.json").is_file()


def test_doc_enum():
    class TestEnum(DocEnum):
        A = "A", "Describes A"
        B = "B", "Might describe B"

    assert str(TestEnum.A) == "A"
    assert TestEnum.B.__doc__ == "Might describe B"



import pytest
from monty.serialization import loadfn
from pymatgen.core import Composition

from emmet.core import ARROW_COMPATIBLE
from emmet.core.structure_group import StructureGroupDoc, _get_id_lexi
from emmet.core.utils import jsanitize

if ARROW_COMPATIBLE:
    import pyarrow as pa

    from emmet.core.arrow import arrowize


@pytest.fixture(scope="session")
def entries_lto(test_dir):
    """
    Recycle the test cases from pymatgen
    """
    entries = loadfn(test_dir / "LiTiO2_batt.json.gz")
    for itr, ient in enumerate(entries):
        ient.entry_id = f"mp-{itr}"
    return entries


@pytest.fixture(scope="session")
def entries_lfeo(test_dir):
    """
    Recycle the test cases from pymatgen
    """
    entries = loadfn(test_dir / "Li-Fe-O.json.gz")
    return entries


def test_StructureGroupDoc_from_grouped_entries(entries_lto):
    sgroup_doc = StructureGroupDoc.from_grouped_entries(
        entries_lto,
        ignored_specie="Li",
    )
    assert sgroup_doc.group_id == "mp-0_Li"
    assert sgroup_doc.material_ids == ["mp-0", "mp-1", "mp-2", "mp-3", "mp-4", "mp-5"]
    assert sgroup_doc.framework_formula == "TiO2"
    assert sgroup_doc.ignored_specie == "Li"
    assert sgroup_doc.chemsys == "Li-O-Ti"
    assert sgroup_doc.has_distinct_compositions is True


@pytest.mark.skip(reason="Temp skip for spglib seg fault")
def test_StructureGroupDoc_from_ungrouped_entries(entries_lfeo):
    entry_dict = {ient.entry_id: ient for ient in entries_lfeo}
    sgroup_docs = StructureGroupDoc.from_ungrouped_structure_entries(
        entries_lfeo, ignored_specie="Li"
    )

    # Make sure that all the structure in each group has the same framework
    for sgroup_doc in sgroup_docs:
        framework_ref = sgroup_doc.framework_formula
        ignored = sgroup_doc.ignored_specie
        for entry_id in sgroup_doc.material_ids:
            dd_ = entry_dict[entry_id].composition.as_dict()
            if ignored in dd_:
                dd_.pop(ignored)
            framework = Composition.from_dict(dd_).reduced_formula
            assert framework == framework_ref


def test_lexi_id():
    assert _get_id_lexi("01HMVV88CCQ6JQ2Y1N8F3ZTVWP") == (
        "01HMVV88CCQ6JQ2Y1N8F3ZTVWP",
        0,
    )
    assert _get_id_lexi("mp-123") == ("mp", 123)
    assert _get_id_lexi("123") == ("", 123)


@pytest.mark.skipif(
    not ARROW_COMPATIBLE, reason="pyarrow must be installed to run this test."
)
def test_arrow(entries_lto):
    doc = StructureGroupDoc.from_grouped_entries(
        entries_lto,
        ignored_specie="Li",
    )
    arrow_struct = pa.scalar(
        doc.model_dump(context={"format": "arrow"}), type=arrowize(StructureGroupDoc)
    )
    test_arrow_doc = StructureGroupDoc(**arrow_struct.as_py(maps_as_pydicts="strict"))

    assert jsanitize(doc.model_dump(), allow_bson=True) == jsanitize(
        test_arrow_doc.model_dump(), allow_bson=True
    )



from __future__ import annotations
from typing import TYPE_CHECKING

from pydantic import BaseModel, Field
from pymatgen.core.structure import Structure

from emmet.core.material_property import PropertyDoc
from emmet.core.utils import generate_robocrys_condensed_struct_and_description

if TYPE_CHECKING:
    from emmet.core.types.typing import IdentifierType


class MineralData(BaseModel):
    """
    Model for mineral data in the condensed structure robocrystallographer field
    """

    type: str | None = Field(
        None,
        description="Mineral type.",
    )

    name: str | None = Field(None, description="The mineral name if found.")


class CondensedStructureData(BaseModel):
    """
    Model for data in the condensed structure robocrystallographer field
    More details: https://hackingmaterials.lbl.gov/robocrystallographer/format.html
    """

    mineral: MineralData = Field(
        description="Matched mineral data for the material.",
    )

    dimensionality: int = Field(
        description="Dimensionality of the material.",
    )

    formula: str | None = Field(
        None,
        description="Formula for the material.",
    )

    spg_symbol: str | None = Field(
        None,
        description="Space group symbol of the material.",
    )

    crystal_system: str | None = Field(
        None,
        description="Crystal system of the material.",
    )


class RobocrystallogapherDoc(PropertyDoc):
    """
    This document contains the descriptive data from robocrystallographer
    for a material:
        Structural features, mineral prototypes, dimensionality, ...

    """

    property_name: str = "robocrys"

    description: str = Field(
        description="Description text from robocrytallographer.",
    )

    condensed_structure: CondensedStructureData = Field(
        description="Condensed structure data from robocrytallographer.",
    )

    robocrys_version: str = Field(
        ...,
        description="The version of Robocrystallographer used to generate this document.",
    )

    @classmethod
    def from_structure(
        cls,
        structure: Structure,
        robocrys_version: str,
        material_id: IdentifierType | None = None,
        mineral_matcher=None,
        **kwargs,
    ):
        (
            condensed_structure,
            description,
        ) = generate_robocrys_condensed_struct_and_description(
            structure=structure, mineral_matcher=mineral_matcher
        )

        return super().from_structure(
            meta_structure=structure,
            material_id=material_id,
            condensed_structure=condensed_structure,
            description=description,
            robocrys_version=robocrys_version,
            **kwargs,
        )



"""Core definition of a Q-Chem Task Document"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from pydantic import BaseModel, Field
from pymatgen.core.structure import Molecule

from emmet.core.mpid import MPID, MPculeID
from emmet.core.qchem.calc_types import (
    CalcType,
    LevelOfTheory,
    TaskType,
    calc_type,
    level_of_theory,
    lot_solvent_string,
    solvent,
    task_type,
)
from emmet.core.structure import MoleculeMetadata
from emmet.core.task import BaseTaskDocument
from emmet.core.types.enums import ValueEnum
from emmet.core.utils import arrow_incompatible

if TYPE_CHECKING:
    from collections.abc import Callable

__author__ = "Evan Spotte-Smith <ewcspottesmith@lbl.gov>"


class QChemStatus(ValueEnum):
    """
    Q-Chem Calculation State
    """

    SUCCESS = "successful"
    FAILED = "unsuccessful"


@arrow_incompatible
class OutputSummary(BaseModel):
    """
    Summary of an output for a Q-Chem calculation
    """

    initial_molecule: Molecule | None = Field(None, description="Input Molecule object")
    optimized_molecule: Molecule | None = Field(
        None, description="Optimized Molecule object"
    )

    final_energy: float | None = Field(
        None, description="Final electronic energy for the calculation (units: Hartree)"
    )
    enthalpy: float | None = Field(
        None, description="Total enthalpy of the molecule (units: kcal/mol)"
    )
    entropy: float | None = Field(
        None, description="Total entropy of the molecule (units: cal/mol-K"
    )

    mulliken: list[Any] | None = Field(
        None, description="Mulliken atomic partial charges and partial spins"
    )
    resp: list[float] | None = Field(
        None,
        description="Restrained Electrostatic Potential (RESP) atomic partial charges",
    )
    nbo: dict[str, Any] | None = Field(
        None, description="Natural Bonding Orbital (NBO) output"
    )

    frequencies: list[float] | None = Field(
        None, description="Vibrational frequencies of the molecule (units: cm^-1)"
    )

    dipoles: dict[str, Any] | None = Field(
        None, description="Electric dipole information for the molecule"
    )

    gradients: list[list[float]] | None = Field(
        None, description="Atomic forces, in atomic units (Ha/Bohr)"
    )

    precise_gradients: list[list[float]] | None = Field(
        None, description="High-precision atomic forces, in atomic units (Ha/Bohr)"
    )

    pcm_gradients: list[list[float]] | None = Field(
        None,
        description="Electrostatic atomic forces from polarizable continuum model (PCM) implicit solvation,"
        "in atomic units (Ha/Bohr).",
    )

    CDS_gradients: list[list[float]] | None = Field(
        None,
        description="Atomic force contributions from cavitation, dispersion, and structural rearrangement in the SMx"
        "family of implicit solvent models, in atomic units (Ha/Bohr)",
    )

    def as_dict(self) -> dict[str, Any]:
        return {
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,
            "initial_molecule": self.initial_molecule,
            "optimized_molecule": self.optimized_molecule,
            "final_energy": self.final_energy,
            "enthalpy": self.enthalpy,
            "entropy": self.entropy,
            "mulliken": self.mulliken,
            "resp": self.resp,
            "nbo": self.nbo,
            "frequencies": self.frequencies,
            "dipoles": self.dipoles,
            "gradients": self.gradients,
            "precise_gradients": self.precise_gradients,
        }


@arrow_incompatible
class TaskDocument(BaseTaskDocument, MoleculeMetadata):
    """
    Definition of a Q-Chem task document
    """

    task_id: MPID | MPculeID | None = Field(
        None, description="the Task ID For this document"
    )

    calc_code: str = "Q-Chem"
    completed: bool = True

    is_valid: bool = Field(
        True, description="Whether this task document passed validation or not"
    )
    state: QChemStatus | None = Field(None, description="State of this calculation")

    cputime: float | None = Field(None, description="The system CPU time in seconds")
    walltime: float | None = Field(None, description="The real elapsed time in seconds")

    calcs_reversed: list[dict] = Field(
        [], description="The 'raw' calculation docs used to assembled this task"
    )

    orig: dict[str, Any] = Field(
        {}, description="Summary of the original Q-Chem inputs"
    )
    output: OutputSummary = Field(OutputSummary())  # type: ignore[call-arg]

    critic2: dict[str, Any] | None = Field(
        None, description="Output from Critic2 critical point analysis code"
    )
    custom_smd: str | None = Field(
        None, description="Parameter string for SMD implicit solvent model"
    )

    special_run_type: str | None = Field(
        None, description="Special workflow name (if applicable)"
    )

    smiles: str | None = Field(
        None,
        description="Simplified molecular-input line-entry system (SMILES) string for the molecule involved "
        "in this calculation.",
    )

    species_hash: str | None = Field(
        None,
        description="Weisfeiler Lehman (WL) graph hash using the atom species as the graph "
        "node attribute.",
    )
    coord_hash: str | None = Field(
        None,
        description="Weisfeiler Lehman (WL) graph hash using the atom coordinates as the graph "
        "node attribute.",
    )

    # TODO - type of `tags` field seems to differ among task databases
    # sometimes List, sometimes Dict
    # left as Any here to ensure tags don't cause validation to fail.
    tags: Any | None = Field(None, description="Metadata tags")

    warnings: dict[str, bool] | None = Field(
        None, description="Any warnings related to this task document"
    )

    @property
    def level_of_theory(self) -> LevelOfTheory:
        return level_of_theory(self.orig)

    @property
    def solvent(self) -> str:
        return solvent(self.orig, custom_smd=self.custom_smd)

    @property
    def lot_solvent(self) -> str:
        return lot_solvent_string(self.orig, custom_smd=self.custom_smd)

    @property
    def task_type(self) -> TaskType:
        return task_type(self.orig, special_run_type=self.special_run_type)

    @property
    def calc_type(self) -> CalcType:
        return calc_type(self.special_run_type, self.orig)

    @property
    def entry(self) -> dict[str, Any]:
        mol = None
        for mol_field in ("optimized_molecule", "initial_molecule"):
            if mol := getattr(self.output, mol_field, None):
                break
        else:
            raise ValueError("No molecule could be associated with the calculation.")

        charge = int(mol.charge) if self.charge is None else int(self.charge)
        spin = (
            mol.spin_multiplicity
            if self.spin_multiplicity is None
            else self.spin_multiplicity
        )

        entry_dict = {
            "entry_id": self.task_id,
            "task_id": self.task_id,
            "charge": charge,
            "spin_multiplicity": spin,
            "level_of_theory": self.level_of_theory,
            "solvent": self.solvent,
            "lot_solvent": self.lot_solvent,
            "custom_smd": self.custom_smd,
            "task_type": self.task_type,
            "calc_type": self.calc_type,
            "molecule": mol,
            "composition": mol.composition,
            "formula": mol.composition.alphabetical_formula,
            "energy": self.output.final_energy,
            "output": self.output.as_dict(),
            "critic2": self.critic2,
            "orig": self.orig,
            "tags": self.tags,
            "last_updated": self.last_updated,
            "species_hash": self.species_hash,
            "coord_hash": self.coord_hash,
        }

        return entry_dict


def filter_task_type(
    entries: list[dict[str, Any]],
    task_type: TaskType,
    sort_by: Callable | None = None,
) -> list[dict[str, Any]]:
    """
    Filter (and sort) TaskDocument entries based on task type

    :param entries: List of TaskDocument entry dicts
    :param TaskType: TaskType to accept
    :param sort_by: Function used to sort (default None)
    :return: Filtered (sorted) list of entries
    """

    filtered = [f for f in entries if f["task_type"] == task_type]
    return sorted(filtered, key=sort_by) if sort_by is not None else filtered



import json

import numpy as np
import pytest
from pymatgen.electronic_structure.core import Spin
from pymatgen.electronic_structure.dos import CompleteDos, Dos
from pymatgen.io.vasp import Vasprun

from emmet.core import ARROW_COMPATIBLE
from emmet.core.band_theory import BaseElectronicDos, ElectronicBS, ElectronicDos
from emmet.core.testing_utils import DataArchive

if ARROW_COMPATIBLE:
    import pyarrow as pa

    from emmet.core.arrow import arrowize


@pytest.fixture(scope="module")
def bs_fixture(test_dir):
    vasprun = DataArchive.extract_obj(
        test_dir / "vasp/r2scan_relax.json.gz",
        "vasprun.xml.gz",
        Vasprun,
        parse_projected_eigen=True,
    )
    pmg_bs = vasprun.get_band_structure()
    elec_bs = ElectronicBS.from_pmg(pmg_bs)

    return (vasprun, pmg_bs, elec_bs)


@pytest.fixture(scope="module")
def dos_fixture(test_dir):
    vasprun = DataArchive.extract_obj(
        test_dir / "vasp/Si_uniform.json.gz",
        "vasprun.xml.gz",
        Vasprun,
        parse_projected_eigen=True,
    )
    pmg_dos = vasprun.complete_dos
    edos = ElectronicDos.from_pmg(pmg_dos)

    return (vasprun, pmg_dos, edos)


def test_elec_band_struct(bs_fixture):
    vasprun, pmg_bs, elec_bs = bs_fixture

    assert len(elec_bs.qpoints) == 116
    assert np.all(
        np.abs(
            np.array(elec_bs.reciprocal_lattice)
            - np.array(vasprun.final_structure.lattice.reciprocal_lattice.matrix)
        )
        < 1e-6
    )
    assert elec_bs.structure == vasprun.final_structure
    assert elec_bs.efermi == pytest.approx(2.72974902)
    assert elec_bs.band_gap == pytest.approx(0.0)
    assert elec_bs.is_metal

    assert all(
        getattr(elec_bs, f"spin_{s}_bands", None) is not None for s in ("up", "down")
    )

    assert all(
        getattr(elec_bs, f"spin_{s}_projections", None) is not None
        for s in ("up", "down")
    )

    dict_rep = elec_bs.model_dump()
    assert isinstance(dict_rep["structure"], dict)

    # round trip JSON
    dumped = elec_bs.model_dump_json()
    assert ElectronicBS(**json.loads(dumped)) == elec_bs

    # round trip pymatgen
    assert pmg_bs.as_dict() == elec_bs.to_pmg().as_dict()


def test_base_electronic_dos():
    """Test base electronic DOS class for jellium."""
    rs = 4.0
    efermi = (9 * np.pi / 4.0) ** (1 / 3) / rs
    energies = np.linspace(0, 3 * efermi, 200)
    heg_dos = (2 * energies) ** (0.5) / np.pi**2

    for spin_pol in (1, 0.5):
        edos = BaseElectronicDos(
            spin_up_densities=spin_pol * heg_dos,
            spin_down_densities=spin_pol * heg_dos if spin_pol < 1 else None,
            energies=energies,
            efermi=efermi,
        )

        spins_to_check = [Spin.up]
        pmg_dos = edos.to_pmg()
        if abs(spin_pol) < 1:
            spins_to_check.append(Spin.down)
        else:
            assert pmg_dos.densities.get(Spin.down) is None

        for spin in spins_to_check:
            assert np.all(
                np.abs(
                    np.array(getattr(edos, f"spin_{spin.name}_densities"))
                    - pmg_dos.densities[spin]
                )
                < 1e-6
            )


def test_electronic_dos(dos_fixture):
    vasprun, pmg_dos, edos = dos_fixture

    assert edos.efermi == pytest.approx(6.26639781)
    assert edos.structure == vasprun.final_structure
    for spin, dos in pmg_dos.densities.items():
        assert np.all(
            np.abs(np.array(getattr(edos, f"spin_{spin.name}_densities")) - dos) < 1e-6
        )
    assert len(edos.projected_densities) == len(edos.structure)

    # test roundtrip json
    dumped = edos.model_dump_json()
    assert ElectronicDos(**json.loads(dumped)) == edos

    # test roundtrip pymatgen
    new_pmg_dos = edos.to_pmg()
    assert new_pmg_dos.as_dict() == pmg_dos.as_dict()
    assert isinstance(new_pmg_dos, CompleteDos)

    # test partial data return type
    edos_incomplete = ElectronicDos(
        **edos.model_dump(exclude=["structure", "projected_densities"])
    )
    assert isinstance(edos_incomplete.to_pmg(), Dos)


@pytest.mark.skipif(
    not ARROW_COMPATIBLE, reason="pyarrow must be installed to run this test."
)
def test_arrow(bs_fixture, dos_fixture):
    _, _, elec_bs = bs_fixture
    bs_arrow_struct = pa.scalar(
        elec_bs.model_dump(context={"format": "arrow"}), type=arrowize(ElectronicBS)
    )
    test_arrow_bs_doc = ElectronicBS(**bs_arrow_struct.as_py(maps_as_pydicts="strict"))
    assert test_arrow_bs_doc == elec_bs

    _, _, edos = dos_fixture
    dos_arrow_struct = pa.scalar(
        edos.model_dump(context={"format": "arrow"}), type=arrowize(ElectronicDos)
    )
    test_arrow_dos_doc = ElectronicDos(
        **dos_arrow_struct.as_py(maps_as_pydicts="strict")
    )
    assert test_arrow_dos_doc == edos



import pytest
from monty.serialization import MontyDecoder, loadfn

from emmet.core import ARROW_COMPATIBLE
from emmet.core.thermo import ThermoDoc

if ARROW_COMPATIBLE:
    import pyarrow as pa

    from emmet.core.arrow import arrowize


@pytest.fixture(scope="session")
def Fe3O4_structure(test_dir):
    structure = loadfn(test_dir / "thermo/Fe3O4_structure.json.gz")
    return structure


@pytest.fixture(scope="session")
def Fe2O3a_structure(test_dir):
    structure = loadfn(test_dir / "thermo/Fe2O3a_structure.json.gz")
    return structure


@pytest.fixture(scope="session")
def Fe2O3b_structure(test_dir):
    structure = loadfn(test_dir / "thermo/Fe2O3b_structure.json.gz")
    return structure


@pytest.fixture(scope="session")
def Fe_structure(test_dir):
    structure = loadfn(test_dir / "thermo/Fe_structure.json.gz")
    return structure


@pytest.fixture(scope="session")
def O_structure(test_dir):
    structure = loadfn(test_dir / "thermo/O_structure.json.gz")
    return structure


@pytest.fixture
def entries(
    Fe3O4_structure, Fe2O3a_structure, Fe2O3b_structure, Fe_structure, O_structure
):
    return MontyDecoder().process_decoded(
        [
            {
                "@module": "pymatgen.entries.computed_entries",
                "@class": "ComputedStructureEntry",
                "correction": 0.0,
                "structure": Fe3O4_structure.as_dict(),
                "entry_id": "mp-1",
                "energy": -382.146593528,
                "composition": {"Fe": 24.0, "O": 32.0},
                "name": "Fe3O4",
                "data": {
                    "material_id": "mp-1",
                    "run_type": "Unknown",
                    "task_id": "mp-10",
                },
                "attribute": None,
                "@version": "2020.4.29",
            },
            {
                "@module": "pymatgen.entries.computed_entries",
                "@class": "ComputedStructureEntry",
                "correction": 0.0,
                "structure": Fe2O3a_structure.as_dict(),
                "entry_id": "mp-2",
                "energy": -270.38765404,
                "composition": {"Fe": 16.0, "O": 24.0},
                "name": "Fe2O3",
                "data": {
                    "material_id": "mp-2",
                    "run_type": "Unknown",
                    "task_id": "mp-20",
                },
                "attribute": None,
                "@version": "2020.4.29",
            },
            {
                "@module": "pymatgen.entries.computed_entries",
                "@class": "ComputedStructureEntry",
                "correction": 0.0,
                "structure": O_structure.as_dict(),
                "entry_id": "mp-3",
                "energy": -92.274692568,
                "composition": {"O": 24.0},
                "name": "O",
                "data": {
                    "material_id": "mp-3",
                    "run_type": "Unknown",
                    "task_id": "mp-30",
                },
                "attribute": None,
                "@version": "2020.4.29",
            },
            {
                "@module": "pymatgen.entries.computed_entries",
                "@class": "ComputedStructureEntry",
                "correction": 0.0,
                "structure": Fe_structure.as_dict(),
                "entry_id": "mp-4",
                "energy": -13.00419661,
                "composition": {"Fe": 2.0},
                "name": "Fe",
                "data": {
                    "material_id": "mp-4",
                    "run_type": "Unknown",
                    "task_id": "mp-40",
                },
                "attribute": None,
                "@version": "2020.4.29",
            },
            {
                "@module": "pymatgen.entries.computed_entries",
                "@class": "ComputedStructureEntry",
                "correction": 0.0,
                "structure": Fe2O3b_structure.as_dict(),
                "entry_id": "mp-5",
                "energy": -1080.82678592,
                "composition": {"Fe": 64.0, "O": 96.0},
                "name": "Fe2O3",
                "attribute": None,
                "data": {
                    "material_id": "mp-5",
                    "run_type": "Unknown",
                    "task_id": "mp-50",
                },
                "@version": "2020.4.29",
            },
        ]
    )


def test_from_entries(entries):
    docs = ThermoDoc.from_entries(entries, thermo_type="UNKNOWN", deprecated=False)

    assert len(docs) == len(entries)

    assert all([d.energy_type == "Unknown" for d in docs])
    unstable_doc = next(d for d in docs if d.material_id == "mp-5")
    assert unstable_doc.is_stable is False
    assert all([d.is_stable for d in docs if d != unstable_doc])


@pytest.mark.skipif(
    not ARROW_COMPATIBLE, reason="pyarrow must be installed to run this test."
)
def test_arrow(entries):
    doc = ThermoDoc.from_entries(entries, thermo_type="UNKNOWN", deprecated=False)[0]

    arrow_struct = pa.scalar(
        doc.model_dump(context={"format": "arrow"}), type=arrowize(ThermoDoc)
    )
    test_arrow_doc = ThermoDoc(**arrow_struct.as_py(maps_as_pydicts="strict"))

    assert doc.model_dump() == test_arrow_doc.model_dump()



"""Define core schemas for VASP calculations."""

from __future__ import annotations

import logging
import re
import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Any

import numpy as np
import orjson
from monty.json import MontyDecoder
from monty.serialization import loadfn
from pydantic import (
    BaseModel,
    Field,
    field_serializer,
    field_validator,
    model_serializer,
    model_validator,
)
from pymatgen.analysis.structure_analyzer import oxide_type
from pymatgen.core.trajectory import Trajectory
from pymatgen.entries.computed_entries import ComputedEntry, ComputedStructureEntry
from pymatgen.io.vasp import Incar, Kpoints, Poscar

from emmet.core.structure import StructureMetadata
from emmet.core.trajectory import Trajectory as CoreTrajectory
from emmet.core.types.enums import TaskState, VaspObject
from emmet.core.types.pymatgen_types.computed_entries_adapter import (
    ComputedEntryType,
    ComputedStructureEntryType,
)
from emmet.core.types.pymatgen_types.structure_adapter import StructureType
from emmet.core.types.pymatgen_types.trajectory_adapter import TrajectoryType
from emmet.core.types.typing import DateTimeType, IdentifierType, NullableDateTimeType
from emmet.core.utils import jsanitize, type_override, utcnow
from emmet.core.vasp.calc_types import (
    CalcType,
    RunType,
    TaskType,
    calc_type,
    run_type,
    task_type,
)
from emmet.core.vasp.calculation import (
    Calculation,
    CalculationInput,
    CoreCalculationOutput,
    PotcarSpec,
    RunStatistics,
)
from emmet.core.vasp.utils import TASK_NAMES, discover_and_sort_vasp_files

if TYPE_CHECKING:
    from collections.abc import Sequence

    from typing_extensions import Self


monty_decoder = MontyDecoder()
logger = logging.getLogger(__name__)

_VOLUMETRIC_FILES = ("CHGCAR", "LOCPOT", "AECCAR0", "AECCAR1", "AECCAR2")


class OrigInputs(CalculationInput):
    """Maintained for backward compatibility - deprecated."""

    def model_post_init(self, __context):
        super().model_post_init(__context)
        warnings.warn(
            f"The `{self.__class__.__name__}` class has been marked "
            "for deprecation to ensure parity of the inputs field "
            "in `TaskDoc`. Please transition to using `CalculationInput`, "
            f"which is fully backwards compatible with `{self.__class__.__name__}`.",
            stacklevel=2,
        )


class InputDoc(OrigInputs):
    """Maintained for backward compatibility - deprecated."""


class OutputDoc(BaseModel):
    structure: StructureType | None = Field(
        None,
        title="Output Structure",
        description="Output Structure from the VASP calculation.",
    )

    density: float | None = Field(None, description="Density of in units of g/cc.")
    energy: float | None = Field(None, description="Total Energy in units of eV.")
    forces: list[list[float]] | None = Field(
        None, description="The force on each atom in units of eV/A."
    )
    stress: list[list[float]] | None = Field(
        None, description="The stress on the cell in units of kB."
    )
    energy_per_atom: float | None = Field(
        None, description="The final DFT energy per atom for the last calculation"
    )
    bandgap: float | None = Field(
        None, description="The DFT bandgap for the last calculation"
    )

    @model_validator(mode="before")
    def set_density_from_structure(cls, values):
        # Validator to automatically set density from structure if not already
        # specified. This might happen when importing an older atomate2-format
        # TaskDocument.
        if not values.get("density", None):
            if isinstance(values["structure"], dict):
                values["density"] = values["structure"].get("density", None)
            else:
                values["density"] = values["structure"].density

        return values

    @classmethod
    def from_vasp_calc_doc(
        cls, calc_doc: Calculation, trajectory: Trajectory | None = None
    ) -> "OutputDoc":
        """
        Create a summary of VASP calculation outputs from a VASP calculation document.

        This will first look for ionic steps in the calculation document.
        If found, will use it and ignore the trajectory.
        If not, will get ionic steps from the trajectory.

        Parameters
        ----------
        calc_doc
            A VASP calculation document.
        trajectory
            A pymatgen Trajectory.

        Returns
        -------
        OutputDoc
            The calculation output summary.
        """
        if calc_doc.output.ionic_steps:
            forces = calc_doc.output.ionic_steps[-1].forces
            stress = calc_doc.output.ionic_steps[-1].stress
        elif trajectory and (ionic_steps := trajectory.frame_properties) is not None:
            forces = ionic_steps[-1].get("forces")
            stress = ionic_steps[-1].get("stress")
        else:
            raise RuntimeError("Unable to find ionic steps.")

        return cls(
            structure=calc_doc.output.structure,
            density=calc_doc.output.structure.density,
            energy=calc_doc.output.energy,
            energy_per_atom=calc_doc.output.energy_per_atom,
            bandgap=calc_doc.output.bandgap,
            forces=forces,
            stress=stress,
        )


@type_override({"corrections": str, "job": str})
class CustodianDoc(BaseModel):
    corrections: list[Any] | None = Field(
        None,
        title="Custodian Corrections",
        description="List of custodian correction data for calculation.",
    )
    job: Any | None = Field(
        None,
        title="Custodian Job Data",
        description="Job data logged by custodian.",
    )

    @model_serializer(mode="wrap")
    def model_serialization(self, default_serializer, info):
        default_serialized_model = default_serializer(self, info)

        format = info.context.get("format") if info.context else None
        if format == "arrow":
            arrow_compat_model = jsanitize(default_serialized_model, allow_bson=True)
            for field, value in arrow_compat_model.items():
                arrow_compat_model[field] = orjson.dumps(value)
            return arrow_compat_model

        return default_serialized_model

    @field_validator("*", mode="before")
    def field_deserializer(cls, field):
        if isinstance(field, str):
            field = orjson.loads(field)
        return field


class AnalysisDoc(BaseModel):
    delta_volume: float | None = Field(
        None,
        title="Volume Change",
        description="Volume change for the calculation.",
    )
    delta_volume_percent: float | None = Field(
        None,
        title="Volume Change Percent",
        description="Percent volume change for the calculation.",
    )
    max_force: float | None = Field(
        None,
        title="Max Force",
        description="Maximum force on any atom at the end of the calculation.",
    )

    warnings: list[str] | None = Field(
        None,
        title="Calculation Warnings",
        description="Warnings issued after analysis.",
    )

    errors: list[str] | None = Field(
        None,
        title="Calculation Errors",
        description="Errors issued after analysis.",
    )

    @classmethod
    def from_vasp_calc_docs(
        cls,
        calcs_reversed: list[Calculation],
        volume_change_warning_tol: float = 0.2,
    ) -> "AnalysisDoc":
        """
        Create analysis summary from VASP calculation documents.

        Parameters
        ----------
        calcs_reversed
            A list of VASP calculation documents in reverse order .
        volume_change_warning_tol
            Maximum volume change allowed in VASP relaxations before the calculation is
            tagged with a warning.

        Returns
        -------
        AnalysisDoc
            The relaxation analysis.
        """
        initial_vol = calcs_reversed[-1].input.structure.lattice.volume
        final_vol = calcs_reversed[0].output.structure.lattice.volume
        delta_vol = final_vol - initial_vol
        percent_delta_vol = 100 * delta_vol / initial_vol
        warnings = []
        errors = []

        if abs(percent_delta_vol) > volume_change_warning_tol * 100:
            warnings.append(f"Volume change > {volume_change_warning_tol * 100}%")

        final_calc = calcs_reversed[0]
        max_force = None
        if final_calc.has_vasp_completed == TaskState.SUCCESS:
            # max force and valid structure checks
            structure = final_calc.output.structure
            # do not check max force for MD run
            if calcs_reversed[0].input.parameters.get("IBRION", -1) != 0:
                max_force = _get_max_force(final_calc)
            warnings.extend(_get_drift_warnings(final_calc))
            if not structure.is_valid():
                errors.append("Bad structure (atoms are too close!)")

        return cls(
            delta_volume=delta_vol,
            delta_volume_percent=percent_delta_vol,
            max_force=max_force,
            warnings=warnings,
            errors=errors,
        )


@type_override({"transformations": str, "vasp_objects": str})
class CoreTaskDoc(StructureMetadata):
    """Calculation-level details about VASP calculations that power the Materials Project."""

    batch_id: str | None = Field(
        None,
        description="Identifier for this calculation; should provide rough information about the calculation origin and purpose.",
    )
    calc_type: CalcType | None = Field(
        None, description="The functional and task type used in the calculation."
    )
    completed_at: NullableDateTimeType = Field(
        description="Timestamp for when this task was completed"
    )
    dir_name: str | None = Field(None, description="The directory for this VASP task")
    icsd_id: int | None = Field(
        None, description="Inorganic Crystal Structure Database id of the structure"
    )
    input: CalculationInput | None = Field(
        None,
        description="VASP calculation inputs",
    )
    last_updated: DateTimeType = Field(
        description="Timestamp for the most recent calculation for this task document",
    )
    orig_inputs: CalculationInput | None = Field(
        None,
        description="The exact set of input parameters used to generate the current task document.",
    )
    output: CoreCalculationOutput | None = Field(
        None,
        description="VASP calculation outputs.",
    )
    run_type: RunType | None = Field(
        None, description="The functional used in the calculation."
    )
    structure: StructureType | None = Field(
        None, description="Final output structure from the task"
    )
    tags: list[str] | None = Field(
        None, title="tag", description="Metadata tagged to a given task."
    )
    task_id: IdentifierType | None = Field(
        None,
        description="The (task) ID of this calculation, used as a universal reference across property documents."
        "This comes in the form: mp-******.",
    )
    task_type: TaskType | CalcType | None = Field(
        None, description="The type of calculation."
    )
    transformations: Any | None = Field(
        None,
        description="Information on the structural transformations, parsed from a "
        "transformations.json file",
    )
    vasp_objects: dict[VaspObject, Any] | None = Field(
        None, description="Vasp objects associated with this task"
    )

    @field_validator("batch_id", mode="before")
    def validate_batch_id(cls, batch_id: str):
        if batch_id:
            invalid_chars = set(
                char
                for char in batch_id
                if (not char.isalnum()) and (char not in {"-", "_"})
            )
            if invalid_chars:
                raise ValueError(
                    f"Invalid characters in batch_id: {' '.join(invalid_chars)}"
                )
        return batch_id

    @field_serializer("transformations", "vasp_objects", mode="wrap")
    def serialize_overrides(self, d, default_serializer, info):
        default_serialized_object = default_serializer(d, info)

        format = info.context.get("format") if info.context else None
        if format == "arrow":
            return orjson.dumps(default_serialized_object)

        return default_serialized_object

    @field_validator("transformations", "vasp_objects", mode="before")
    def deserialize_overrides(cls, d):
        return orjson.loads(d) if isinstance(d, str) else d

    @classmethod
    def from_directory(
        cls,
        dir_name: Path | str,
        volumetric_files: tuple[str, ...] = _VOLUMETRIC_FILES,
        **vasp_calculation_kwargs,
    ) -> tuple[Self, CoreTrajectory]:
        """
        Create a CoreTaskDoc and corresponding CoreTrajectory from a
        directory containing VASP files.

        Only parses a single calculation. Use TaskDoc.from_directory(...)
        to parse multiple calculations.

        Parameters
        ----------
        dir_name
            The path to the folder containing the calculation outputs.
        volumetric_files
            Volumetric files to search for.
        **vasp_calculation_kwargs
            Additional parsing options that will be passed to the
            :obj:`.Calculation.from_vasp_files` function.

        Returns
        -------
        CoreTaskDoc
            A slim task document for the calculation.
        CoreTrajectory
            A low memory document for the calculation's trajectory.
        """
        dir_name = Path(dir_name)
        task_files = _find_vasp_files(dir_name, volumetric_files=volumetric_files)

        calc_doc, vasp_objects = Calculation.from_vasp_files(
            dir_name, "standard", **task_files["standard"], **vasp_calculation_kwargs
        )
        transformations, icsd_id, tags, author = _parse_transformations(dir_name)
        task_doc = cls.from_structure(
            calc_type=calc_doc.calc_type,
            completed_at=calc_doc.completed_at,
            dir_name=get_uri(dir_name),
            icsd_id=icsd_id,
            input=InputDoc.from_vasp_calc_doc(calc_doc),
            meta_structure=calc_doc.output.structure,
            orig_inputs=_parse_orig_inputs(dir_name),
            output=CoreCalculationOutput(
                **calc_doc.output.model_dump(
                    include=set(CoreCalculationOutput.model_fields())
                )
            ),
            run_type=calc_doc.run_type,
            structure=calc_doc.output.structure,
            tags=tags,
            task_type=calc_doc.task_type,
            transformations=transformations,
            vasp_objects=vasp_objects,
        )

        props: dict[str, Any] = {
            "structure": [],
            "cart_coords": [],
            "electronic_steps": [],
            "energy": [],
            "e_wo_entrp": [],
            "e_fr_energy": [],
            "forces": [],
            "stress": [],
        }

        for ionic_step in calc_doc.output.ionic_steps:
            props["structure"].append(ionic_step.structure)
            props["energy"].append(ionic_step.e_0_energy)
            for k in (
                "e_fr_energy",
                "e_wo_entrp",
                "forces",
                "stress",
                "electronic_steps",
            ):
                props[k].append(getattr(ionic_step, k))

        trajectory = CoreTrajectory._from_dict(
            props, task_type=calc_doc.task_type, run_type=calc_doc.run_type
        )

        return (task_doc, trajectory)


@type_override({"additional_json": str})
class TaskDoc(CoreTaskDoc, extra="allow"):
    """Flexible wrapper around CoreTaskDoc"""

    additional_json: dict[str, Any] | None = Field(
        None, description="Additional json loaded from the calculation directory"
    )
    analysis: AnalysisDoc | None = Field(
        None,
        title="Calculation Analysis",
        description="Some analysis of calculation data after collection.",
    )
    author: str | None = Field(
        None, description="Author extracted from transformations"
    )
    calcs_reversed: list[Calculation] | None = Field(
        None,
        title="Calcs reversed data",
        description="Detailed data for each VASP calculation contributing to the task document.",
    )
    custodian: list[CustodianDoc] | None = Field(
        None,
        title="Calcs reversed data",
        description="Detailed custodian data for each VASP calculation contributing to the task document.",
    )
    entry: ComputedEntryType | None = Field(
        None, description="The ComputedEntry from the task doc"
    )
    included_objects: list[VaspObject] | None = Field(
        None, description="List of VASP objects included with this task document"
    )
    output: OutputDoc | None = Field(
        None,
        description="The exact set of output parameters used to generate the current task document.",
    )
    run_stats: dict[str, RunStatistics] | None = Field(
        None,
        description="Summary of runtime statistics for each calculation in this task",
    )
    state: TaskState | None = Field(None, description="State of this calculation")
    task_label: str | None = Field(None, description="A description of the task")

    @model_validator(mode="before")
    @classmethod
    def set_model_pre_fields(cls, values: Any) -> Any:
        """Ensure all important model fields are set and refreshed."""

        # Always refresh task_type, calc_type, run_type
        # if attributes containing input sets are available.
        # See, e.g. https://github.com/materialsproject/emmet/issues/960
        # where run_type's were set incorrectly in older versions of TaskDoc
        attrs = ["calcs_reversed", "input", "orig_inputs"]
        for icalc, calc in enumerate(values.get("calcs_reversed", [])):
            if isinstance(calc, dict):
                values["calcs_reversed"][icalc] = Calculation(**calc)

        calcs_reversed = values.get("calcs_reversed")

        if any(values.get(attr) is not None for attr in attrs):
            # To determine task and run type, we search for input sets in this order
            # of precedence: calcs_reversed, inputs, orig_inputs
            inp_set = None
            inp_sets_to_check = [values.get("input"), values.get("orig_inputs")]
            if calcs_reversed:
                inp_sets_to_check = [calcs_reversed[0].get("input")] + inp_sets_to_check

            for inp_set in inp_sets_to_check:
                if inp_set is not None:
                    values["task_type"] = task_type(inp_set)
                    break

            # calcs_reversed needed below
            if calcs_reversed:
                values["run_type"] = cls._get_run_type(calcs_reversed)
                if inp_set is not None:
                    values["calc_type"] = cls._get_calc_type(calcs_reversed, inp_set)

        if calcs_reversed:
            # TODO: remove after imposing TaskDoc schema on older tasks in collection
            if final_struct := calcs_reversed[0].output.structure:
                values["structure"] = values.get("structure", final_struct)
                values["entry"] = values.get(
                    "entry", cls.get_entry(calcs_reversed, values.get("task_id"))
                )

        return values

    @field_serializer("additional_json", mode="wrap")
    def serialize_additional_json(self, d, default_serializer, info):
        default_serialized_object = default_serializer(d, info)

        format = info.context.get("format") if info.context else None
        if format == "arrow":
            return orjson.dumps(default_serialized_object)

        return default_serialized_object

    @field_validator("additional_json", mode="before")
    def deserialize_additional_json(cls, d):
        return orjson.loads(d) if isinstance(d, str) else d

    @classmethod
    def from_directory(
        cls,
        dir_name: Path | str,
        volumetric_files: tuple[str, ...] = _VOLUMETRIC_FILES,
        store_additional_json: bool = True,
        additional_fields: dict[str, Any] | None = None,
        volume_change_warning_tol: float = 0.2,
        task_names: list[str] | None = None,
        **vasp_calculation_kwargs,
    ) -> Self:
        """
        Create a task document from a directory containing VASP files.


        Parameters
        ----------
        dir_name
            The path to the folder containing the calculation outputs.
        store_additional_json
            Whether to store additional json files found in the calculation directory.
        volumetric_files
            Volumetric files to search for.
        additional_fields
            Dictionary of additional fields to add to output document.
        volume_change_warning_tol
            Maximum volume change allowed in VASP relaxations before the calculation is
            tagged with a warning.
        task_names
            Naming scheme for multiple calculations in on folder e.g. ["relax1","relax2"].
            Can be subfolder or extension.
        **vasp_calculation_kwargs
            Additional parsing options that will be passed to the
            :obj:`.Calculation.from_vasp_files` function.

        Returns
        -------
        TaskDoc
            A task document for the calculation.
        """
        logger.info(f"Getting task doc in: {dir_name}")

        additional_fields = {} if additional_fields is None else additional_fields
        dir_name = Path(dir_name).resolve()
        task_files = _find_vasp_files(
            dir_name, volumetric_files=volumetric_files, task_names=task_names
        )

        if len(task_files) == 0:
            raise FileNotFoundError("No VASP files found!")

        calcs_reversed = []
        all_vasp_objects = []
        for task_name in sorted(task_files):
            calc_doc, vasp_objects = Calculation.from_vasp_files(
                dir_name, task_name, **task_files[task_name], **vasp_calculation_kwargs
            )
            calcs_reversed.append(calc_doc)
            all_vasp_objects.append(vasp_objects)

        # Reverse the list of calculations in the order:  newest calc is the first
        # To match with calcs_reversed, all_vasp_objects is also reversed.
        calcs_reversed.reverse()
        all_vasp_objects.reverse()

        analysis = AnalysisDoc.from_vasp_calc_docs(
            calcs_reversed, volume_change_warning_tol=volume_change_warning_tol
        )
        transformations, icsd_id, tags, author = _parse_transformations(dir_name)
        custodian = _parse_custodian(dir_name)
        orig_inputs = _parse_orig_inputs(dir_name)

        additional_json = None
        if store_additional_json:
            additional_json = _parse_additional_json(dir_name)

        dir_name = get_uri(dir_name)  # convert to full uri path

        # only store objects from last calculation
        # TODO: make this an option
        vasp_objects = all_vasp_objects[0]
        included_objects = None
        if vasp_objects:
            included_objects = list(vasp_objects.keys())

        doc = cls.from_structure(
            structure=calcs_reversed[0].output.structure,
            meta_structure=calcs_reversed[0].output.structure,
            dir_name=dir_name,
            calcs_reversed=calcs_reversed,
            analysis=analysis,
            transformations=transformations,
            custodian=custodian,
            orig_inputs=orig_inputs,
            additional_json=additional_json,
            icsd_id=icsd_id,
            tags=tags,
            author=author,
            completed_at=calcs_reversed[0].completed_at,
            input=calcs_reversed[-1].input,
            output=OutputDoc.from_vasp_calc_doc(
                calcs_reversed[0],
                vasp_objects.get(VaspObject.TRAJECTORY),  # type: ignore
            ),
            state=_get_state(calcs_reversed, analysis),
            run_stats=_get_run_stats(calcs_reversed),
            vasp_objects=vasp_objects,
            included_objects=included_objects,
            task_type=calcs_reversed[0].task_type,
        )
        return doc.model_copy(update=additional_fields)

    @classmethod
    def from_vasprun(
        cls,
        path: str | Path,
        additional_fields: dict[str, Any] | None = None,
        volume_change_warning_tol: float = 0.2,
        **vasp_calculation_kwargs,
    ) -> Self:
        """
        Create a task document from a vasprun.xml file.

        This is not recommended and will raise warnings, since some necessary
        information is absent from the vasprun.xml file, such as per-atom
        magnetic moments. However, the majority of the TaskDoc will
        be complete.

        Parameters
        ----------
        path
            The path to the vasprun.xml.
        additional_fields: dict[str, Any] = None,
        volume_change_warning_tol
            Maximum volume change allowed in VASP relaxations before the calculation is
            tagged with a warning.
        **vasp_calculation_kwargs
            Additional parsing options that will be passed to the
            :obj:`.Calculation.from_vasp_files` function.

        Returns
        -------
        TaskDoc
            A task document for the calculation.
        """
        logger.info(f"Getting vasprun.xml at: {path}")

        path = Path(path)
        dir_name = path.resolve().parent

        calc = Calculation.from_vasprun(path, **vasp_calculation_kwargs)
        calcs_reversed = [calc]

        analysis = AnalysisDoc.from_vasp_calc_docs(
            calcs_reversed, volume_change_warning_tol=volume_change_warning_tol
        )

        # assume orig_inputs are those stated in vasprun.xml
        orig_inputs = CalculationInput(
            incar=calc.input.incar,
            structure=calc.input.structure,
            kpoints=calc.input.kpoints,
            potcar=calc.input.potcar,
        )

        doc = cls.from_structure(
            structure=calcs_reversed[0].output.structure,
            meta_structure=calcs_reversed[0].output.structure,
            dir_name=get_uri(dir_name),
            calcs_reversed=calcs_reversed,
            analysis=analysis,
            orig_inputs=orig_inputs,
            completed_at=calcs_reversed[0].completed_at,
            input=calcs_reversed[-1].input,
            output=OutputDoc.from_vasp_calc_doc(calcs_reversed[0]),
            state=_get_state(calcs_reversed, analysis),
            run_stats=None,
            vasp_objects={},
            included_objects=[],
            task_type=calcs_reversed[0].task_type,
        )
        if additional_fields:
            doc = doc.model_copy(update=additional_fields)
        return doc

    @staticmethod
    def get_entry(
        calcs_reversed: list[Calculation | dict],
        task_id: IdentifierType | str | int | None = None,
    ) -> ComputedEntry:
        """
        Get a computed entry from a list of VASP calculation documents.

        Parameters
        ----------
        calcs_reversed
            A list of VASP calculation documents in a reverse order.
        task_id
            The job identifier.

        Returns
        -------
        ComputedEntry
            A computed entry.
        """
        if isinstance(cr := calcs_reversed[0], dict):
            cr = Calculation(**cr)
        calc_inp = cr.input
        calc_out = cr.output

        entry_dict = {
            "correction": 0.0,
            "entry_id": task_id,
            "composition": calc_out.structure.composition,
            "energy": calc_out.energy,
            "parameters": {
                # Cannot be PotcarSpec document, pymatgen expects a dict
                # Note that `potcar_spec` is optional
                "potcar_spec": (
                    [dict(d) for d in calc_inp.potcar_spec]
                    if calc_inp.potcar_spec
                    else []
                ),
                # Required to be compatible with MontyEncoder for the ComputedEntry
                "run_type": str(cr.run_type),
                "is_hubbard": calc_inp.is_hubbard,
                "hubbards": calc_inp.hubbards,
            },
            "data": {
                "oxide_type": oxide_type(calc_out.structure),
                "aspherical": calc_inp.parameters.get("LASPH", False),
                "last_updated": str(utcnow()),
            },
        }
        return ComputedEntry.from_dict(entry_dict)

    @staticmethod
    def _get_calc_type(
        calcs_reversed: list[Calculation], orig_inputs: CalculationInput
    ) -> CalcType:
        """Get the calc type from calcs_reversed.

        Returns
        --------
        CalcType
            The type of calculation.
        """
        if isinstance(calcs_reversed[0], Calculation):
            cr_inp = calcs_reversed[0].input
            params = cr_inp.parameters
            incar = cr_inp.incar
        else:
            cr_inp = calcs_reversed[0].get("input", {})
            params = cr_inp.get("parameters", {})
            incar = cr_inp.get("incar", {})

        inputs = cr_inp if len(calcs_reversed) > 0 else orig_inputs
        return calc_type(inputs, {**params, **incar})

    @staticmethod
    def _get_run_type(calcs_reversed: list[Calculation | dict]) -> RunType:
        """Get the run type from calcs_reversed.

        Returns
        --------
        RunType
            The type of calculation.
        """
        if isinstance(calcs_reversed[0], Calculation):
            params = calcs_reversed[0].input.parameters
            incar = calcs_reversed[0].input.incar
        else:
            cr_inp = calcs_reversed[0].get("input", {})
            params = cr_inp.get("parameters", {})
            incar = cr_inp.get("incar", {})
        return run_type({**params, **incar})

    @property
    def structure_entry(self) -> ComputedStructureEntry:
        """
        Retrieve a ComputedStructureEntry for this TaskDoc.

        Returns
        -------
        ComputedStructureEntry
            The TaskDoc.entry with corresponding TaskDoc.structure added.
        """
        if not self.structure or not self.entry:
            raise ValueError(
                "Need both a `structure` and `entry` to return a `ComputedStructureEntry`."
            )

        return ComputedStructureEntry(
            structure=self.structure,
            energy=self.entry.energy,
            correction=self.entry.correction,
            composition=self.entry.composition,
            energy_adjustments=self.entry.energy_adjustments,
            parameters=self.entry.parameters,
            data=self.entry.data,
            entry_id=self.entry.entry_id,
        )


class TrajectoryDoc(BaseModel):
    """Model for task trajectory data."""

    task_id: str | None = Field(
        None,
        description="The (task) ID of this calculation, used as a universal reference across property documents."
        "This comes in the form: mp-******.",
    )

    trajectories: list[TrajectoryType] | None = Field(
        None,
        description="Trajectory data for calculations associated with a task doc.",
    )


class EntryDoc(BaseModel):
    """Model for task entry data."""

    task_id: str | None = Field(
        None,
        description="The (task) ID of this calculation, used as a universal reference across property documents."
        "This comes in the form: mp-******.",
    )

    entry: ComputedStructureEntryType | None = Field(
        None,
        description="Computed structure entry for the calculation associated with the task doc.",
    )


class DeprecationDoc(BaseModel):
    """Model for task deprecation data."""

    task_id: str | None = Field(
        None,
        description="The (task) ID of this calculation, used as a universal reference across property documents."
        "This comes in the form: mp-******.",
    )

    deprecated: bool | None = Field(
        None,
        description="Whether the ID corresponds to a deprecated calculation.",
    )

    deprecation_reason: str | None = Field(
        None,
        description="Reason for deprecation.",
    )


def get_uri(dir_name: str | Path) -> str:
    """
    Return the URI path for a directory.

    This allows files hosted on different file servers to have distinct locations.

    Parameters
    ----------
    dir_name : str or Path
        A directory name.

    Returns
    -------
    str
        Full URI path, e.g., "fileserver.host.com:/full/path/of/dir_name".
    """
    import socket

    fullpath = Path(dir_name).absolute()
    hostname = socket.gethostname()
    try:
        hostname = socket.gethostbyaddr(hostname)[0]
    except (socket.gaierror, socket.herror):
        pass
    return f"{hostname}:{fullpath}"


def _parse_transformations(
    dir_name: Path,
) -> tuple[dict, int | None, list[str] | None, str | None]:
    """Parse transformations.json file."""
    transformations = {}
    filenames = tuple(dir_name.glob("transformations.json*"))
    icsd_id = None
    if len(filenames) >= 1:
        transformations = loadfn(filenames[0], cls=None)
        try:
            match = re.match(r"(\d+)-ICSD", transformations["history"][0]["source"])
            if match:
                icsd_id = int(match.group(1))
        except (KeyError, IndexError):
            pass

    # We don't want to leave tags or authors in the
    # transformations file because they'd be copied into
    # every structure generated after this one.
    other_parameters = transformations.get("other_parameters", {})
    new_tags = other_parameters.pop("tags", None)
    new_author = other_parameters.pop("author", None)

    if "other_parameters" in transformations and not other_parameters:
        # if dict is now empty remove it
        transformations.pop("other_parameters")

    return transformations, icsd_id, new_tags, new_author


def _parse_custodian(dir_name: Path) -> dict | None:
    """
    Parse custodian.json file.

    Calculations done using custodian have a custodian.json file which tracks the makers
    performed and any errors detected and fixed.

    Parameters
    ----------
    dir_name
        Path to calculation directory.

    Returns
    -------
    dict | None
        The information parsed from custodian.json file.
    """
    filenames = tuple(dir_name.glob("custodian.json*"))
    if len(filenames) >= 1:
        return loadfn(filenames[0], cls=None)
    return None


def _parse_orig_inputs(
    dir_name: Path, suffix: str | None = ".orig"
) -> dict[str, Kpoints | Poscar | PotcarSpec | Incar]:
    """
    Parse original input files.

    Calculations using custodian generate a *.orig file for the inputs. This is useful
    to know how the calculation originally started.

    Parameters
    ----------
    dir_name
        Path to calculation directory.
    suffix : str or None = ".orig"
        The suffix of the original input files to use.

    Returns
    -------
    dict[str, Kpoints | Poscar | PotcarSpec | Incar ]
        The original POSCAR, KPOINTS, POTCAR, and INCAR data.
    """
    orig_inputs = {}
    input_mapping: dict[str, Kpoints | Poscar | PotcarSpec | Incar] = {
        "INCAR": Incar,
        "KPOINTS": Kpoints,
        "POTCAR": PotcarSpec,
        "POSCAR": Poscar,
    }
    suffix = suffix or ""
    for filename in dir_name.glob("*".join(f"{suffix}.".split("."))):
        if f"POTCAR.spec{suffix}" in str(filename):
            try:
                orig_inputs["potcar_spec"] = PotcarSpec.from_file(filename)
            except Exception:
                # Can't parse non emmet-core style POTCAR.spec files
                continue
        for name, vasp_input in input_mapping.items():
            if f"{name}{suffix}" in str(filename):
                file_suffix = "_spec" if name == "POTCAR" else ""
                orig_inputs[f"{name.lower()}{file_suffix}"] = vasp_input.from_file(
                    filename
                )

    return orig_inputs


def _parse_additional_json(dir_name: Path) -> dict[str, Any]:
    """Parse additional json files in the directory."""
    additional_json = {}
    for filename in dir_name.glob("*.json*"):
        key = filename.name.split(".")[0]
        # ignore FW.json(.gz) so jobflow doesn't try to parse prev_vasp_dir OutputReferences
        # was causing atomate2 MP workflows to fail with ValueError: Could not resolve reference
        # 7f5a7f14-464c-4a5b-85f9-8d11b595be3b not in store or cache
        # contact @janosh in case of questions
        if key not in ("custodian", "transformations", "FW"):
            additional_json[key] = loadfn(filename, cls=None)
    return additional_json


def _get_max_force(calc_doc: Calculation) -> float | None:
    """Get max force acting on atoms from a calculation document."""
    if calc_doc.output.ionic_steps:
        forces: np.ndarray | list | None = None
        if calc_doc.output.ionic_steps:
            forces = calc_doc.output.ionic_steps[-1].forces

        structure = calc_doc.output.structure
        if forces:
            forces = np.array(forces)
            sdyn = structure.site_properties.get("selective_dynamics")
            if sdyn:
                forces[np.logical_not(sdyn)] = 0
            return max(np.linalg.norm(forces, axis=1))
    return None


def _get_drift_warnings(calc_doc: Calculation) -> list[str]:
    """Get warnings of whether the drift on atoms is too large."""
    warnings = []
    if calc_doc.input.parameters.get("NSW", 0) > 0:
        drift = calc_doc.output.outcar.get("drift", [[0, 0, 0]])
        max_drift = max(np.linalg.norm(d) for d in drift)  # type: ignore[type-var]
        ediffg = calc_doc.input.parameters.get("EDIFFG", None)
        max_force = -float(ediffg) if ediffg and float(ediffg) < 0 else np.inf
        if max_drift > max_force:
            warnings.append(
                f"Drift ({drift}) > desired force convergence ({max_force}), structure "
                "likely not converged to desired accuracy."
            )
    return warnings


def _get_state(calcs_reversed: list[Calculation], analysis: AnalysisDoc) -> TaskState:
    """Get state from calculation documents and relaxation analysis."""
    all_calcs_completed = all(
        c.has_vasp_completed == TaskState.SUCCESS for c in calcs_reversed
    )
    if (
        analysis.errors is None
        or (isinstance(analysis.errors, list) and len(analysis.errors) == 0)
    ) and all_calcs_completed:
        return TaskState.SUCCESS  # type: ignore
    return TaskState.FAILED  # type: ignore


def _get_run_stats(calcs_reversed: list[Calculation]) -> dict[str, RunStatistics]:
    """Get summary of runtime statistics for each calculation in this task."""
    run_stats = {}
    total = dict(
        average_memory=0.0,
        max_memory=0.0,
        elapsed_time=0.0,
        system_time=0.0,
        user_time=0.0,
        total_time=0.0,
        cores=0,
    )
    for calc_doc in calcs_reversed:
        stats = calc_doc.output.run_stats
        run_stats[calc_doc.task_name] = stats
        total["average_memory"] = max(total["average_memory"], stats.average_memory)
        total["max_memory"] = max(total["max_memory"], stats.max_memory)
        total["cores"] = max(total["cores"], stats.cores)
        total["elapsed_time"] += stats.elapsed_time
        total["system_time"] += stats.system_time
        total["user_time"] += stats.user_time
        total["total_time"] += stats.total_time
    run_stats["overall"] = RunStatistics(**total)
    return run_stats


def _find_vasp_files(
    path: str | Path,
    volumetric_files: Sequence[str] | None = None,
    task_names: Sequence[str] | None = None,
) -> dict[str, dict[str, Path | list[Path]]]:
    """
    Find VASP files in a directory.

    Only files in folders with names matching a task name (or alternatively files
    with the task name as an extension, e.g., vasprun.relax1.xml) will be returned.

    VASP files in the current directory will be given the task name "standard".

    Parameters
    ----------
    path
        Path to a directory to search.
    volumetric_files : sequence of str, defaults to _VOLUMETRIC_FILES
        Volumetric files to search for.
    task_names : sequence of str, defaults to
        task_names = ["precondition","standard","relax0","relax1",..."relax8"]


    Returns
    -------
    dict[str,dict[str,Path | list[Path]]]
        The filenames of the calculation outputs for each VASP task, given as a ordered
        dictionary of::

            {
                task_name: {
                    "vasprun_file": vasprun_filename,
                    "outcar_file": outcar_filename,
                    "contcar_file": contcar_filename,
                    "potcar_spec_file": potcar_spec_filename,
                    "volumetric_files": [CHGCAR, LOCPOT, etc]
                    "elph_poscars": [POSCAR.T=300, POSCAR.T=400, etc]
                },
                ...
            }
    """
    base_path = Path(path)
    volumetric_files = volumetric_files or _VOLUMETRIC_FILES
    task_names = task_names or TASK_NAMES

    task_files: dict[str, dict[str, Path | list[Path]]] = discover_and_sort_vasp_files(
        base_path
    )
    # TODO: TaskDoc permits matching sub directories if they use one of
    # `task_names` as a directory name.
    # Not sure this is behavior we want to keep in the long term,
    # but is maintained here for backwards compatibility.
    for task_name in set(task_names).difference(task_files):
        if (subdir := base_path / task_name).exists():
            for task_name, calcs in discover_and_sort_vasp_files(subdir).items():
                task_files[task_name].update(calcs)

    return task_files



"""Define schemas for DFPT, phonopy, and pheasy-derived phonon data."""

from __future__ import annotations

from enum import Enum
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
import yaml  # type: ignore[import-untyped]
from monty.io import zopen
from monty.os.path import zpath
from pydantic import (
    BaseModel,
    Field,
    PrivateAttr,
    computed_field,
    field_serializer,
    field_validator,
    model_validator,
)
from pymatgen.core import Lattice, Structure
from pymatgen.electronic_structure.bandstructure import BandStructureSymmLine, Kpoint
from pymatgen.electronic_structure.core import Spin
from pymatgen.phonon.bandstructure import PhononBandStructureSymmLine
from pymatgen.phonon.dos import CompletePhononDos
from pymatgen.phonon.dos import PhononDos as PhononDosObject
from typing_extensions import Literal, TypedDict

from emmet.core.band_theory import BandStructure, BandTheoryBase
from emmet.core.base import CalcMeta
from emmet.core.math import Matrix3D, Tensor4R, Vector3D
from emmet.core.polar import BornEffectiveCharges, DielectricDoc, IRDielectric
from emmet.core.structure import StructureMetadata
from emmet.core.types.enums import DocEnum
from emmet.core.types.pymatgen_types.structure_adapter import StructureType
from emmet.core.types.typing import DateTimeType, FSPathType, IdentifierType
from emmet.core.utils import get_num_formula_units, type_override

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Any

    from typing_extensions import Self


DEFAULT_PHONON_FILES = {
    "structure": "POSCAR",
    "phonon_bandstructure": "band.yaml",  # chosen as Phonopy default
    "phonon_dos": "total_dos.dat",  # chosen as Phonopy default
    "force_constants": "FORCE_CONSTANTS",  # chosen as Phonopy default
    "born": "born.npz",
    "epsilon_static": "epsilon_static.npz",
    "phonopy_output": "phonopy.yaml",
}


class PhononMethod(Enum):
    """Define common methods for computed phonon properties."""

    DFPT = "dfpt"
    PHONOPY = "phonopy"
    PHEASY = "pheasy"


class PhononDOS(BandTheoryBase):
    """Define schema of pymatgen phonon density of states."""

    frequencies: list[float] = Field(description="The phonon frequencies in THz.")
    densities: list[float] = Field(description="The phonon density of states.")
    projected_densities: list[list[float]] | None = Field(
        None, description="The projected phonon density of states."
    )

    @classmethod
    def from_pmg(cls, config: PhononDosObject | CompletePhononDos | dict) -> Any:
        """Correctly parse pymatgen-like objects."""
        remap = {
            "structure": "structure",
            "frequencies": "frequencies",
            "densities": "densities",
            "projected_densities": "pdos",
        }
        if isinstance(config, PhononDosObject | CompletePhononDos):
            cls_config = {k: getattr(config, v, None) for k, v in remap.items()}
            if isinstance(pdos_as_dict := cls_config.get("projected_densities"), dict):
                if not cls_config["structure"]:
                    raise ValueError(
                        "Cannot parse atom-projected phonon density of states without a structure."
                    )
                cls_config["projected_densities"] = [
                    list(pdos_as_dict.get(site, [])) for site in cls_config["structure"]
                ]
            return cls(**cls_config)
        return cls(**{k: config.get(v) for k, v in remap.items()})

    @cached_property
    def to_pmg(self) -> PhononDosObject | CompletePhononDos:
        """Get / cache corresponding pymatgen object."""
        dos = PhononDosObject(frequencies=self.frequencies, densities=self.densities)
        if self.projected_densities and self.structure:
            return CompletePhononDos(
                self.structure,
                dos,
                dict(zip(self.structure, self.projected_densities, strict=True)),
            )
        return dos

    @classmethod
    def from_phonopy(
        cls,
        phonon_dos_file: FSPathType,
    ) -> Self:
        """Create a PhononDOS from phonopy .dat output.

        Parameters
        -----------
        phonon_dos_file (FSPathType) : path to total_dos.dat
        """
        phonopy_dos: dict[str, Any] = {
            k: []
            for k in (
                "frequencies",
                "densities",
            )
        }
        with zopen(phonon_dos_file, "rt") as f:
            for line in f.read().splitlines():
                non_comment_text = line.split("#")[0]  # type: ignore[arg-type]
                if len(cols := non_comment_text.split()) == 2:
                    phonopy_dos["frequencies"].append(float(cols[0]))
                    phonopy_dos["densities"].append(float(cols[1]))
                elif len(cols) > 2:
                    raise ValueError(
                        f"File {phonon_dos_file} does not have the correct "
                        "phonopy total_dos.dat format."
                    )

        return cls(**phonopy_dos)


class ShreddedEigendisplacements(TypedDict):
    real: list[list[list[tuple[float, float, float]]]]
    imag: list[list[list[tuple[float, float, float]]]]


@type_override({"eigendisplacements": ShreddedEigendisplacements})
class PhononBS(BandStructure):
    """Define schema of pymatgen phonon band structure."""

    has_nac: bool = Field(
        False,
        description="Whether the calculation includes non-analytical corrections at Gamma.",
    )

    frequencies: list[list[float]] = Field(
        description="The eigen-frequencies, with the first index representing the band, and the second the k-point.",
        validation_alias="bands",
    )

    eigendisplacements: list[list[list[tuple[complex, complex, complex]]]] | None = (
        Field(None, description="Phonon eigendisplacements in Cartesian coordinates.")
    )

    _primitive_structure: Structure | None = PrivateAttr(None)

    @field_serializer("eigendisplacements", mode="wrap")
    def eigendisplacements_serializer(self, eigen, default_serializer, info):
        arr = np.array(default_serializer(eigen, info))
        return {
            "real": arr.real.tolist(),
            "imag": arr.imag.tolist(),
        }

    @field_validator("eigendisplacements", mode="before")
    def eigendisplacements_deserializer(cls, eigen):
        if isinstance(eigen, dict) and all(
            eigen.get(k) is not None for k in ("real", "imag")
        ):
            real = np.array(eigen["real"])
            imag = np.array(eigen["imag"])
            return real + 1.0j * imag
        return eigen

    @classmethod
    def from_pmg(cls, config: PhononBandStructureSymmLine | dict) -> Any:
        """Ensure fields are correctly populated."""
        config = (
            config.as_dict()
            if isinstance(config, PhononBandStructureSymmLine)
            else config
        )

        # legacy data contains abipy structure objects
        if (struct := config.get("structure")) and not isinstance(struct, Structure):
            config["structure"] = Structure.from_dict(struct)

        # remap legacy fields
        for k, v in {
            "lattice_rec": "reciprocal_lattice",
            "bands": "frequencies",
        }.items():
            if config.get(k):
                config[v] = config.pop(k)

        if isinstance(config["reciprocal_lattice"], dict):
            config["reciprocal_lattice"] = config["reciprocal_lattice"].get("matrix")
        return cls(**config)

    @property
    def primitive_structure(self) -> Structure | None:
        """Cache primitive structure for use in computing entropy, heat capacity, etc."""
        if self.structure and not self._primitive_structure:
            self._primitive_structure = self.structure.get_primitive_structure()
        return self._primitive_structure

    @cached_property
    def to_pmg(self) -> PhononBandStructureSymmLine:
        """Get / cache corresponding pymatgen object."""
        rlatt = Lattice(self.reciprocal_lattice)
        return PhononBandStructureSymmLine(
            [Kpoint(q, lattice=rlatt).frac_coords for q in self.qpoints],  # type: ignore[misc]
            np.array(self.frequencies),
            rlatt,
            has_nac=self.has_nac,
            eigendisplacements=np.array(self.eigendisplacements),
            structure=self.structure,
            labels_dict={
                k: Kpoint(v, lattice=rlatt).frac_coords
                for k, v in (self.labels_dict or {}).items()
            },
            coords_are_cartesian=False,
        )

    @property
    def _to_pmg_es_bs(self) -> BandStructureSymmLine:
        """Convenience method for crystal toolkit to create electronic-structure style band structure object."""
        rlatt = Lattice(self.reciprocal_lattice)
        return BandStructureSymmLine(
            [Kpoint(q, lattice=rlatt).frac_coords for q in self.qpoints],  # type: ignore[arg-type]
            {Spin.up: np.array(self.frequencies)},  # type: ignore[dict-item]
            rlatt,
            1e-6,  # There is no Fermi level in a Phonon DOS (these are bosons definitionally) but we want to plot the bands from zero.
            {
                k: Kpoint(v, lattice=rlatt).frac_coords  # type: ignore[misc]
                for k, v in (self.labels_dict or {}).items()
            },
            coords_are_cartesian=False,
            structure=self.structure,
        )

    @classmethod
    def from_phonopy(cls, phonon_bandstructure_file: FSPathType):
        """Create a PhononBS from phonopy .yaml output."""
        with zopen(phonon_bandstructure_file, "rt") as f:
            phonopy_bandstructure = yaml.safe_load(f.read())

        phonopy_bandstructure.update(
            qpoints=[entry["q-position"] for entry in phonopy_bandstructure["phonon"]],
            frequencies=[
                [branch["frequency"] for branch in entry["band"]]
                for entry in phonopy_bandstructure["phonon"]
            ],
        )
        return cls(**phonopy_bandstructure)


class SumRuleChecks(BaseModel):
    """Container class for defining sum rule checks."""

    asr: float | None = Field(
        None, description="The violation of the acoustic sum rule."
    )
    cnsr: float | None = Field(
        None, description="The violation of the charge neutral sum rule."
    )


class PhononComputationalSettings(BaseModel):
    """Collection to store computational settings for the phonon computation."""

    # could be optional and implemented at a later stage?
    npoints_band: int | None = Field(
        None, description="number of points for band structure computation"
    )
    kpath_scheme: str | None = Field(None, description="indicates the kpath scheme")
    kpoint_density_dos: int | None = Field(
        None,
        description="number of points for computation of free energies and densities of states",
    )


class ThermalDisplacementData(BaseModel):
    """Collection to store information on the thermal displacement matrices."""

    freq_min_thermal_displacements: float | None = Field(
        None,
        description="cutoff frequency in THz to avoid numerical issues in the "
        "computation of the thermal displacement parameters",
    )
    thermal_displacement_matrix_cif: list[list[Matrix3D]] | None = Field(
        None, description="field including thermal displacement matrices in CIF format"
    )
    thermal_displacement_matrix: list[list[Matrix3D]] | None = Field(
        None,
        description="field including thermal displacement matrices in Cartesian "
        "coordinate system",
    )
    temperatures_thermal_displacements: list[float] | None = Field(
        None,
        description="temperatures at which the thermal displacement matrices"
        "have been computed",
    )


class PhononBSDOSTask(StructureMetadata):
    """Phonon band structures and density of states data."""

    identifier: str | None = Field(
        None, description="The identifier of this phonon analysis task."
    )

    phonon_method: PhononMethod | None = Field(
        None, description="The method used to calculate phonon properties."
    )

    phonon_bandstructure: PhononBS | None = Field(
        None,
        description="Phonon band structure object.",
    )

    phonon_dos: PhononDOS | None = Field(
        None,
        description="Phonon density of states object.",
    )

    epsilon_static: Matrix3D | None = Field(
        None, description="The high-frequency dielectric constant."
    )

    epsilon_electronic: Matrix3D | None = Field(
        None,
        description="The electronic contribution to the high-frequency dielectric constant.",
    )

    born: list[Matrix3D] | None = Field(
        None,
        description="Born charges, only for symmetrically inequivalent atoms",
    )

    # needed, e.g. to compute Grueneisen parameter etc
    force_constants: list[list[Matrix3D]] | None = Field(
        None, description="Force constants between every pair of atoms in the structure"
    )

    last_updated: DateTimeType = Field(
        description="Timestamp for the most recent calculation for this Material document.",
    )

    sum_rules_breaking: SumRuleChecks | None = Field(
        None,
        description="Deviations from sum rules.",
    )

    structure: StructureType | None = Field(
        None, description="Structure used in the calculation."
    )

    total_dft_energy: float | None = Field(
        None, description="total DFT energy in eV/atom."
    )

    volume_per_formula_unit: float | None = Field(
        None, description="volume per formula unit in Angstrom**3."
    )

    formula_units: int | None = Field(None, description="Formula units per cell.")

    supercell_matrix: Matrix3D | None = Field(
        None, description="matrix describing the supercell."
    )
    primitive_matrix: Matrix3D | None = Field(
        None, description="matrix describing relationship to primitive cell."
    )

    code: str | None = Field(
        None, description="String describing the code for the computation."
    )

    post_process_settings: PhononComputationalSettings | None = Field(
        None,
        description="Field including settings for the post processing code, e.g., phonopy.",
    )

    thermal_displacement_data: ThermalDisplacementData | None = Field(
        None,
        description="Includes all data of the computation of the thermal displacements",
    )

    calc_meta: list[CalcMeta] | None = Field(
        None,
        description="Metadata for individual calculations used to build this document.",
    )

    @classmethod
    def migrate_fields(cls, **config) -> Any:
        """Migrate legacy input fields."""

        # This block is for older DFPT data
        for k, v in {
            "ph_dos": "phonon_dos",
            "ph_bs": "phonon_bandstructure",
            "e_total": "epsilon_static",
            "e_electronic": "epsilon_electronic",
            "becs": "born",
        }.items():
            if config.get(k):
                config[v] = config.pop(k)

        # migrate pymatgen objects
        for k, emmet_cls in {
            "phonon_dos": PhononDOS,
            "phonon_bandstructure": PhononBS,
        }.items():
            if (old_obj := config.get(k)) and not isinstance(old_obj, emmet_cls):
                try:
                    new_obj = emmet_cls.from_pmg(old_obj)  # type: ignore[attr-defined]
                except Exception:
                    # purposely no except here - want this to hard fail if we
                    # can't parse the input type
                    new_obj = emmet_cls(**old_obj)
                config[k] = new_obj

        if (ph_bs := config.get("phonon_bandstructure")) and not config.get(
            "structure"
        ):
            config["structure"] = ph_bs.structure

        # This block is for the migration from atomate2 --> emmet-core schemas
        if config.get("structure"):
            if isinstance(config["structure"], dict):
                config["structure"] = Structure.from_dict(config["structure"])

            old_formula_units = config.pop("formula_units", 1)
            if toten := config.pop("total_dft_energy", None):
                config["total_dft_energy"] = (
                    toten * old_formula_units / config["structure"].num_sites
                )

            config["formula_units"] = get_num_formula_units(
                config["structure"].composition
            )

        if (fc := config.get("force_constants")) and isinstance(fc, dict):
            config["force_constants"] = fc["force_constants"]

        if comp_sett := config.get("phonopy_settings"):
            config["post_process_settings"] = comp_sett

        calc_meta_migrate = {
            "uuids": "_uuid",
            "jobdirs": "_job_dir",
        }
        calc_meta_remap = {"uuids": "identifier", "jobdirs": "dir_name"}
        if any(config.get(k) for k in calc_meta_migrate):
            calc_meta: dict[str, dict[str, str]] = {}
            for field_name, splitter in calc_meta_migrate.items():
                if vals := config.get(field_name):
                    if not isinstance(vals, dict):
                        vals = vals.model_dump()
                    for k, v in vals.items():
                        job_name = k.split(splitter)[0]
                        if isinstance(v, list):
                            for idx, sub_v in enumerate(v):
                                if (sub_name := f"{job_name}-{idx}") not in calc_meta:
                                    calc_meta[sub_name] = {}
                                calc_meta[sub_name][calc_meta_remap[field_name]] = sub_v
                        else:
                            if job_name not in calc_meta:
                                calc_meta[job_name] = {}
                            calc_meta[job_name][calc_meta_remap[field_name]] = v

            config["calc_meta"] = [
                CalcMeta(
                    name=job_name,
                    **meta,
                )
                for job_name, meta in calc_meta.items()
            ]

        return cls.from_structure(
            meta_structure=config.pop("meta_structure", config.get("structure")),
            **config,
        )

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def has_imaginary_modes(self) -> bool | None:
        tol: float = 1e-5
        if self.phonon_bandstructure:
            return self.phonon_bandstructure.to_pmg.has_imaginary_freq(tol=tol)
        elif self.phonon_dos:
            return bool(
                np.any(
                    np.array(self.phonon_dos.densities)[
                        np.array(self.phonon_dos.frequencies) < tol
                    ]
                    > tol
                )
            )
        return None

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def charge_neutral_sum_rule(self) -> Matrix3D | None:
        """Sum of Born effective charges over sites should be zero."""
        if self.born:
            bec = np.array(self.born)
            return tuple(tuple(row) for row in np.sum(bec, axis=0).tolist())
        return None

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def acoustic_sum_rule(self) -> Matrix3D | None:
        """Sum of q=0 atomic force constants should be zero."""
        if self.force_constants:
            return tuple(
                tuple(row)
                for row in np.einsum("iijk->jk", np.array(self.force_constants))
            )
        return None

    @computed_field  # type: ignore[prop-decorator]
    @property
    def check_sum_rule_deviations(self) -> SumRuleChecks:
        """Report deviations from sum rules."""
        if not self.sum_rules_breaking:
            self.sum_rules_breaking = SumRuleChecks(
                **{
                    k: (
                        np.max(np.abs(getattr(self, attr)))
                        if getattr(self, attr)
                        else None
                    )
                    for k, attr in {
                        "asr": "acoustic_sum_rule",
                        "cnsr": "charge_neutral_sum_rule",
                    }.items()
                }
            )
        return self.sum_rules_breaking

    def _get_thermo_from_dos(
        self,
        quantity: Literal["entropy", "cv", "internal_energy", "helmholtz_free_energy"],
        temperature: float,
        normalization: Literal["atoms", "formula_units"] | None = "formula_units",
    ) -> float | None:
        """Get a thermodynamic property from the phonon DOS.

        quantity : "entropy", "cv", "internal_energy", or "helmholtz_free_energy"
            If "entropy" or "cv", the units before normalization are J/(K * mol).
            If "internal_energy" or "helmholtz_free_energy", the units before normalization are J/mol.
        temperature : float
        normalization : "atoms", "formula_units", or None
            Whether to normalize by the number of atoms in the cell ("atoms"),
            the number of formula units ("formula_units", default), or not (None).
        """

        if not self.phonon_dos:
            return None

        if normalization and not self.structure:
            raise ValueError(
                "Cannot normalize thermodynamic quantities without a structure."
            )

        norm_fac = 1.0
        if normalization == "atoms":
            norm_fac = 1.0 / self.structure.num_sites  # type: ignore[union-attr]
        elif normalization == "formula_units":
            norm_fac = 1.0 / self.formula_units  # type: ignore[operator]
        elif normalization:
            raise ValueError(f"Unknown {normalization=} convention.")

        return (
            getattr(self.phonon_dos.to_pmg, quantity)(
                temp=temperature,
            )
            * norm_fac
        )

    def entropy(
        self,
        temperature: float,
        normalization: Literal["atoms", "formula_units"] | None = "formula_units",
    ) -> float | None:
        """Compute the entropy in J/(K * mol * formula units).

        temperature : float
        normalization : "atoms", "formula_units", or None
            Whether to normalize by the number of atoms in the cell ("atoms"),
            the number of formula units ("formula_units"), or not (None).
        """
        return self._get_thermo_from_dos(
            "entropy", temperature, normalization=normalization
        )

    def heat_capacity(
        self,
        temperature: float,
        normalization: Literal["atoms", "formula_units"] | None = "formula_units",
    ) -> float | None:
        """Compute the heat capacity in J/(K * mol * formula units).

        temperature : float
        normalization : "atoms", "formula_units", or None
            Whether to normalize by the number of atoms in the cell ("atoms"),
            the number of formula units ("formula_units"), or not (None).
        """
        return self._get_thermo_from_dos("cv", temperature, normalization=normalization)

    def internal_energy(
        self,
        temperature: float,
        normalization: Literal["atoms", "formula_units"] | None = "formula_units",
    ) -> float | None:
        """Compute the internal energy in J/(mol * formula units).

        temperature : float
        normalization : "atoms", "formula_units", or None
            Whether to normalize by the number of atoms in the cell ("atoms"),
            the number of formula units ("formula_units"), or not (None).
        """
        return self._get_thermo_from_dos(
            "internal_energy", temperature, normalization=normalization
        )

    def free_energy(
        self,
        temperature: float,
        normalization: Literal["atoms", "formula_units"] | None = "formula_units",
    ) -> float | None:
        """Compute the Helmholtz free energy in J/(mol * formula units).

        temperature : float
        normalization : "atoms", "formula_units", or None
            Whether to normalize by the number of atoms in the cell ("atoms"),
            the number of formula units ("formula_units"), or not (None).
        """
        return self._get_thermo_from_dos(
            "helmholtz_free_energy", temperature, normalization=normalization
        )

    def compute_thermo_quantities(
        self,
        temperatures: Sequence[float],
        normalization: Literal["atoms", "formula_units"] | None = "formula_units",
    ) -> dict[str, Sequence[float | None]]:
        """Compute all thermodynamic quantities as a convenience method."""

        quantities = {
            "entropy": "entropy",
            "heat_capacity": "cv",
            "internal_energy": "internal_energy",
            "free_energy": "helmholtz_free_energy",
        }

        thermo_props: dict[str, Sequence[float | None]] = {
            k: [
                self._get_thermo_from_dos(v, temp, normalization=normalization)  # type: ignore[arg-type]
                for temp in temperatures
            ]
            for k, v in quantities.items()
        }
        thermo_props["temperature"] = temperatures
        return thermo_props

    @classmethod
    def from_phonopy_pheasy_files(
        cls,
        structure_file: FSPathType,
        phonon_bandstructure_file: FSPathType | None = None,
        phonon_dos_file: FSPathType | None = None,
        force_constants_file: FSPathType | None = None,
        born_file: FSPathType | None = None,
        epsilon_static_file: FSPathType | None = None,
        phonopy_output_file: FSPathType | None = None,
        **kwargs,
    ) -> Self:
        """
        Create a PhononBSDOSDoc from a list of explicit Phonopy/Pheasy file paths.
        """

        cls_config: dict[str, Any] = {
            "structure": Structure.from_file(structure_file),
        }
        if "poscar" in str(structure_file).lower():
            cls_config["code"] = "vasp"

        if phonon_bandstructure_file:
            cls_config["phonon_bandstructure"] = PhononBS.from_phonopy(
                phonon_bandstructure_file
            )

        if phonon_dos_file:
            cls_config["phonon_dos"] = PhononDOS.from_phonopy(phonon_dos_file)

        if force_constants_file:
            # read FORCE_CONSTANTS manually
            force_constant_matrix: np.ndarray
            idxs: tuple[int | None, int | None] = (
                None,
                None,
            )
            irow = 0
            with zopen(force_constants_file, "rt") as f:
                for idx, line in enumerate(f.read().splitlines()):
                    vals = line.strip().split()
                    if idx == 0:
                        force_constant_matrix = np.zeros(
                            (int(vals[0]), int(vals[1]), 3, 3)
                        )
                    elif len(vals) == 2:
                        # idxs written like this for mypy
                        idxs = (
                            int(vals[0]) - 1,
                            int(vals[1]) - 1,
                        )
                        irow = 0
                    elif len(vals) == 3:
                        force_constant_matrix[idxs[0], idxs[1], irow] = [
                            float(v) for v in vals
                        ]
                        irow += 1
            cls_config["force_constants"] = force_constant_matrix.tolist()

        if born_file:
            cls_config["born"] = np.load(born_file)
        if epsilon_static_file:
            cls_config["epsilon_static"] = np.load(epsilon_static_file)

        if phonopy_output_file:
            with zopen(phonopy_output_file, "rt") as f:
                phonopy_output = yaml.safe_load(f.read())
            for k in ("primitive_matrix", "supercell_matrix"):
                cls_config[k] = phonopy_output.get(k)

        return cls.from_structure(cls_config["structure"], **cls_config, **kwargs)

    @classmethod
    def from_phonopy_pheasy_directory(
        cls,
        phonon_dir: Path | str,
        **kwargs,
    ) -> Self:
        """Create a PhononBSDOSDoc from a Phonopy/Pheasy directory.

        Parameters
        -----------
        phonon_dir : str or Path
        **kwargs to pass to `PhononBSDOSDoc.from_phonopy_pheasy_files`
        """
        phonon_path = Path(phonon_dir).resolve()
        file_paths = {}
        for k, file_name in DEFAULT_PHONON_FILES.items():
            if (file_path := Path(zpath(str(phonon_path / file_name)))).exists():
                file_paths[f"{k}_file"] = file_path
        return cls.from_phonopy_pheasy_files(**file_paths, **kwargs)


class PhononBSDOSDoc(PhononBSDOSTask):
    """Built data version of PhononBSDOSTask."""

    material_id: IdentifierType | None = Field(
        None,
        description="The Materials Project ID of the material, of the form mp-******.",
    )
    task_ids: list[str] | None = Field(
        None, description="A list of identifiers that were used to build this document."
    )

    @model_validator(mode="after")
    def match_id_fields(self) -> Self:
        """Ensure that `material_id` aliases inherited `identifier` field."""
        self.identifier = self.material_id
        return self


class PhononWarnings(DocEnum):
    ASR = "ASR break", "acoustic sum rule max breaking is larger than 30 cm^-1."
    CNSR = "CNSR break", "charge neutrality sum rule max breaking is larger than 0.2."
    NEG_FREQ = (
        "has negative frequencies",
        "phonon band structure has negative "
        "frequencies anywhere in the Brillouin zone.",
    )
    SMALL_Q_NEG_FREQ = (
        "has small q negative frequencies",
        "the phonon band structure has negative frequencies,"
        " but these are small and very close to the Gamma point "
        "(usually related to numerical errors).",
    )


vec2D = list[float, float]  # type: ignore[type-arg]
vec3D = list[float, float, float]  # type: ignore[type-arg]


class PhononWebsiteStruct(TypedDict):
    atom_numbers: list[int]
    atom_pos_car: list[vec3D]
    atom_pos_red: list[vec3D]
    atom_types: list[str]
    distances: list[float]
    eigenvalues: list[list[float]]
    formula: str
    # fails arrow conversion
    # highsym_qpts: list[list[int, str]]
    lattice: list[vec3D]
    line_breaks: list[tuple[int, int]]
    name: str
    natoms: int
    qpoints: list[vec3D]
    repetitions: list[int, int, int]  # type: ignore[type-arg]
    vectors: list[list[list[list[vec2D, vec2D, vec2D]]]]  # type: ignore[type-arg]


class PhononWebsiteBS(BaseModel):
    """
    Document with a serialized version of the phonon band structure suitable
    for the phononwebsite (http://henriquemiranda.github.io/phononwebsite/).
    """

    material_id: str = Field(
        ...,
        description="The ID of this material, used as a universal reference across property documents."
        "This comes in the form: mp-******",
    )

    doc_type: Literal["phononwebsite"] = Field(
        "phononwebsite",
        description="The type of the document: a phonon band structure for the phononwebsite.",
    )

    phononwebsite: PhononWebsiteStruct | None = Field(
        None,
        description="Phononwebsite dictionary to plot the animated " "phonon modes.",
    )

    last_updated: DateTimeType = Field(
        description="Timestamp for the most recent calculation update for this property",
    )

    created_at: DateTimeType = Field(
        description="Timestamp for when this material document was first created",
    )


class Ddb(BaseModel):
    """
    Document with a the string version of the DDB file produced by abinit.
    """

    material_id: str = Field(
        ...,
        description="The ID of this material, used as a universal reference across property documents."
        "This comes in the form: mp-******",
    )

    doc_type: Literal["ddb"] = Field(
        "ddb", description="The type of the document: a DDB file."
    )

    ddb: str | None = Field(None, description="The string of the DDB file.")

    last_updated: DateTimeType = Field(
        description="Timestamp for the most recent calculation update for this property",
    )

    created_at: DateTimeType = Field(
        description="Timestamp for when this material document was first created",
    )


class ThermodynamicProperties(BaseModel):
    """
    Definition of the thermodynamic properties extracted from the phonon frequencies.
    """

    temperatures: list[float] = Field(
        ...,
        description="The list of temperatures at which the thermodynamic properties "
        "are calculated",
    )

    cv: list[float] = Field(
        ...,
        description="The values of the constant-volume specific heat.",
        alias="heat_capacity",
    )

    entropy: list[float] = Field(
        ..., description="The values of the vibrational entropy."
    )


class VibrationalEnergy(BaseModel):
    """
    Definition of the vibrational contribution to the energy as function of
    the temperature.
    """

    temperatures: list[float] = Field(
        ...,
        description="The list of temperatures at which the thermodynamic properties "
        "are calculated",
    )

    internal_energy: list[float] = Field(
        ..., description="The values of the phonon contribution to the internal energy."
    )

    helmholtz_free_energy: list[float] = Field(
        ..., description="The values of the Helmholtz free energy."
    )

    zero_point_energy: float = Field(
        ..., description="The value of the zero point energy."
    )


class Phonon(StructureMetadata):
    """
    Definition for a document with data produced by a phonon calculation.
    """

    material_id: str = Field(
        ...,
        description="The ID of this material, used as a universal reference across property documents."
        "This comes in the form: mp-******",
    )

    structure: StructureType = Field(
        ..., description="The relaxed structure for the phonon calculation."
    )

    asr_break: float | None = Field(
        None, description="The maximum breaking of the acoustic sum rule (ASR)."
    )

    warnings: list[PhononWarnings] | None = Field(
        None, description="List of warnings associated to the phonon calculation."
    )

    dielectric: DielectricDoc | None = Field(
        None, description="Dielectric properties obtained during a phonon calculations."
    )

    becs: BornEffectiveCharges | None = Field(
        None, description="Born effective charges obtained for a phonon calculation."
    )

    ir_spectra: IRDielectric | None = Field(None, description="The IRDielectricTensor.")

    thermodynamic: ThermodynamicProperties | None = Field(
        None,
        description="The thermodynamic properties extracted from the phonon "
        "frequencies.",
    )

    vibrational_energy: VibrationalEnergy | None = Field(
        None, description="The vibrational contributions to the total energy."
    )

    last_updated: DateTimeType = Field(
        description="Timestamp for when this document was last updated",
    )

    created_at: DateTimeType = Field(
        description="Timestamp for when this material document was first created",
    )


class TypedBaseInputDict(TypedDict):
    charge: float
    chksymbreak: int
    ecut: float
    fband: float
    kptopt: int
    nband: int
    nbdbuf: int
    ngkpt: list[int]
    nshiftk: int
    nspden: int
    nspinor: int
    nsppol: int
    nstep: int
    pawecutdg: float
    shiftk: list[list[float]]
    tolvrs: float


class TypedDdBaseDict(TypedBaseInputDict):
    chkprim: int
    nqpt: int
    qpt: list[int]
    rfdir: list[int]
    rfelfd: int


class TypedDdeInputDict(TypedDdBaseDict):
    prtwf: int


class TypedDdkInputDict(TypedDdBaseDict):
    iscf: float


class TypedPhononInputDict(TypedDdeInputDict):
    rfatpol: list[int]


class TypedPseudopotentialsDict(TypedDict):
    md5: list[str]
    name: list[str]


class TypedAbinitInputVars(TypedDict):
    dde_input: TypedDdeInputDict
    ddk_input: TypedDdkInputDict
    ecut: float
    gs_input: TypedBaseInputDict
    ngkpt: list[int]
    ngqpt: list[int]
    occopt: int
    phonon_input: TypedPhononInputDict
    pseudopotentials: TypedPseudopotentialsDict
    shiftk: list[list[float]]
    tsmear: int
    wfq_input: TypedBaseInputDict  # have to guess here without an example entry


class AbinitPhonon(Phonon):
    """
    Definition for a document with data produced from a phonon calculation
    with Abinit.
    """

    abinit_input_vars: TypedAbinitInputVars | None = Field(
        None,
        description="Dict representation of the inputs used to obtain the phonon"
        "properties and the main general options (e.g. number of "
        "k-points, number of q-points).",
    )


class SoundVelocity(BaseModel):
    """
    Definition for a document with the sound velocities of the acoustic modes
    close to Gamma, as obtained from a phonon calculation.
    """

    material_id: str = Field(
        ...,
        description="The ID of this material, in the form: mp-******",
    )

    structure: StructureType = Field(
        ..., description="The relaxed structure for the phonon calculation."
    )

    directions: list[Vector3D] = Field(
        ...,
        description="Q-points identifying the directions for the calculation"
        "of the speed of sound. In fractional coordinates.",
    )

    labels: list[str | None] = Field(..., description="labels of the directions.")

    sound_velocities: list[Vector3D] = Field(
        ...,
        description="Values of the sound velocities in SI units.",
    )

    mode_types: list[tuple[str | None, str | None, str | None]] = Field(
        ...,
        description="The types of the modes ('transversal', 'longitudinal'). "
        "None if not correctly identified.",
    )

    last_updated: DateTimeType = Field(
        description="Timestamp for when this document was last updated",
    )

    created_at: DateTimeType = Field(
        description="Timestamp for when this material document was first created",
    )


class ThermalDisplacement(BaseModel):
    """
    Definition of a Document for the generalized density of states and
    mean square displacements related to phonon oscillations.
    """

    material_id: str = Field(
        ...,
        description="The ID of this material, used as a universal reference across property documents."
        "This comes in the form: mp-******",
    )

    last_updated: DateTimeType = Field(
        description="Timestamp for the most recent calculation update for this property",
    )

    created_at: DateTimeType = Field(
        description="Timestamp for when this material document was first created",
    )

    nsites: int = Field(
        ...,
        description="The number of sites in the structure.",
    )

    nomega: int = Field(
        ...,
        description="The number of frequencies.",
    )

    ntemp: int = Field(
        ...,
        description="The number of temperatures for which the displacements are calculated",
    )

    temperatures: list[float] = Field(
        ...,
        description="The list of temperatures at which the thermodynamic properties "
        "are calculated",
    )

    frequencies: list[float] = Field(
        ..., description="The list of frequencies for the generalized DOS"
    )

    gdos_aijw: Tensor4R = Field(
        ...,
        description=" Generalized DOS in Cartesian coords, with shape (nsites, 3, 3, nomega)",
    )

    amu: dict[str, float] = Field(
        ..., description="Dictionary of the atomic masses in atomic units."
    )

    structure: StructureType = Field(
        ..., description="The relaxed structure for the phonon calculation."
    )

    ucif_t: Tensor4R = Field(
        ...,
        description="Mean squared displacement U tensors as a function of T for T in tmesh in CIF format."
        "With shape (natom, 3, 3, ntemp) ",
    )
    ucif_string_t300k: str = Field(
        ...,
        description="Mean squared displacement U tensors at T=300K in CIF string format.",
    )



"""Test structure similarity features."""

import pytest

from pymatgen.core import Structure
import numpy as np

from emmet.core.similarity import (
    CrystalNNSimilarity,
    M3GNetSimilarity,
    CrystalNNFingerprint,
    matgl,
    vector_difference_matrix,
    SimilarityScorer,
    SimilarityDoc,
    SimilarityEntry,
)


structures = {
    "fcc_cu": Structure(
        3.5 * np.array([[0.0, 0.5, 0.5], [0.5, 0.0, 0.5], [0.5, 0.5, 0.0]]),
        ["Cu"],
        [[0.0, 0.0, 0.0]],
    ),  # fcc Cu
    "dc_cu_cl": Structure(
        4 * np.array([[0.0, 0.5, 0.5], [0.5, 0.0, 0.5], [0.5, 0.5, 0.0]]),
        ["Cu", "Cl"],
        [[0.0, 0.0, 0.0], [0.25, 0.25, 0.25]],
    ),  # diamond cubic CuCl
    "hcp_mo_s": Structure(
        2.8
        * np.array(
            [
                [0.5, -(3.0 ** (0.5)) / 2.0, 0.0],
                [0.5, 3.0 ** (0.5) / 2.0, 0.0],
                [0.0, 0.0, 2],
            ]
        ),
        ["Mo", "S"],
        [
            [2 / 3, 1 / 3, 0.25],
            [1 / 3, 2 / 3, 0.75],
        ],
    ),  # weird hcp MoS
}


@pytest.mark.skipif(CrystalNNFingerprint is None, reason="matminer is not installed.")
def test_crystalnn_featurize():

    scorer = CrystalNNSimilarity()
    feature_vectors, dist_matrix = scorer.get_all_similarity_scores(
        list(structures.values())
    )
    assert feature_vectors.shape == (
        len(structures),
        122,
    )  # this majik number is defined by the fingerprint

    ref_matrix = np.array(
        [
            [100.0, 5.27097101, 35.0290187],
            [5.27097101, 100.0, 8.01686918],
            [35.0290187, 8.01686918, 100.0],
        ]
    )
    assert dist_matrix.shape == (len(structures), len(structures))
    assert np.all(np.abs(dist_matrix - dist_matrix.T) < 1e-6)  # should be symmetric
    assert np.all(np.abs(dist_matrix - ref_matrix) < 1e-6)

    num_matches = 2
    similarity_docs = scorer.build_similarity_collection_from_structures(
        structures,
        num_procs=1,
        num_top=num_matches,
    )

    # Ensure that similarity docs are correctly constructed,
    # and order of docs follows order of structures in the
    # source dict.
    # Also check that the number of included similarity entries
    # is the same as specified by `num_matches`
    assert all(
        (
            isinstance(similarity_docs[i], SimilarityDoc)
            and len(similarity_docs[i].sim) == num_matches
            and similarity_docs[i].material_id == idx
            and all(
                isinstance(entry, SimilarityEntry) for entry in similarity_docs[i].sim
            )
        )
        for i, idx in enumerate(structures.keys())
    )

    # Check that the feature vectors are the same as before
    assert np.all(
        np.abs(fv - similarity_docs[i].feature_vector) < 1e-6
        for i, fv in enumerate(feature_vectors)
    )

    # Check that the matched similarity entries have the same
    # similarity scores as computed in `ref_matrix`
    id_to_index = {idfr: i for i, idfr in enumerate(structures.keys())}
    assert all(
        all(
            100 - entry.dissimilarity
            == pytest.approx(ref_matrix[i, id_to_index[entry.task_id]])
            for entry in doc.sim
        )
        for i, doc in enumerate(similarity_docs)
    )


@pytest.mark.skipif(matgl is None, reason="matgl is not installed.")
def test_m3gnet_featurize():

    scorer = M3GNetSimilarity()
    fvs, dist_matrix = scorer.get_all_similarity_scores(list(structures.values()))

    assert fvs.shape == (len(structures), 128)

    ref_matrix = np.array(
        [
            [100.0, 3.274432636807312, 0.1410324007577568],
            [3.274432636807312, 100.0, 0.20956578539322868],
            [0.1410324007577568, 0.20956578539322868, 100.0],
        ]
    )

    assert np.all(np.abs(dist_matrix - dist_matrix.T) < 1e-6)  # should be symmetric
    assert np.all(np.abs(dist_matrix - ref_matrix) < 1e-6)


def test_vector_diff_mat():

    vectors = np.random.random((7, 11))
    brute_force_diffs = np.zeros(2 * (vectors.shape[0],))
    for i in range(vectors.shape[0]):
        for j in range(vectors.shape[0]):
            brute_force_diffs[i, j] = np.linalg.norm(vectors[i] - vectors[j])
    assert np.all(np.abs(brute_force_diffs - vector_difference_matrix(vectors)) < 1e-12)


def test_vendi():

    assert all(
        abs(rank - SimilarityScorer.get_vendi_score(np.eye(rank))) < 1e-12
        for rank in range(1, 6)
    )

    assert all(
        abs(1.0 - SimilarityScorer.get_vendi_score(np.ones((rank, rank)))) < 1e-12
        for rank in range(1, 6)
    )



from datetime import datetime

from pydantic import BaseModel, Field

from emmet.core.openmm.tasks import Calculation


class CalculationsDoc(BaseModel):
    """
    A document for storing metadata from a list of OpenMM calculations.

    In each field, calculations are listed sequentially, in the order they were run.
    """

    task_names: list[str] | None = Field(None, description="Names of tasks.")

    calc_types: list[str] | None = Field(None, description="Types of calculations.")

    elapsed_times: list[float | None] | None = Field(
        None, description="Elapsed time for calculations."
    )

    steps: list[float | None] | None = Field(
        None, description="n_steps for calculations."
    )

    step_sizes: list[float | None] | None = Field(
        None, description="Step sizes for each calculations."
    )

    temperatures: list[float | None] | None = Field(
        None, description="Temperature for each calculations."
    )

    pressures: list[float | None] | None = Field(
        None, description="Pressure for each calculations."
    )

    friction_coefficients: list[float | None] | None = Field(
        None,
        description="Friction coefficients for each calculations.",
    )

    completed_at: datetime | None = Field(
        None,
        description="Timestamp for when the final calculation completed.",
    )

    job_uuid: str | None = Field(
        None, description="The UUID of the flow that generated this data."
    )

    flow_uuid: str | None = Field(
        None, description="The UUID of the top level host from that job."
    )

    @classmethod
    def from_calcs_reversed(
        cls,
        calcs_reversed: list[Calculation],
        job_uuid: str | None = None,
        flow_uuid: str | None = None,
    ) -> "CalculationsDoc":
        calcs = calcs_reversed[::-1]
        return CalculationsDoc(
            task_names=[calc.task_name for calc in calcs],
            calc_types=[calc.calc_type for calc in calcs],
            elapsed_times=[calc.output.elapsed_time for calc in calcs],
            steps=[calc.input.n_steps for calc in calcs],
            step_sizes=[calc.input.step_size for calc in calcs],
            temperatures=[calc.input.temperature for calc in calcs],
            pressures=[calc.input.pressure for calc in calcs],
            friction_coefficients=[calc.input.friction_coefficient for calc in calcs],
            completed_at=calcs[-1].completed_at,
            job_uuid=job_uuid,
            flow_uuid=flow_uuid,
        )



"""Schemas for classical MD package."""

from __future__ import annotations

import zlib
from datetime import datetime

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    PlainSerializer,
    PlainValidator,
    WithJsonSchema,
)
from pymatgen.core import Structure
from typing_extensions import Annotated

from emmet.core.utils import arrow_incompatible, utcnow
from emmet.core.vasp.task_valid import TaskState


def compressed_str_validator(s: str) -> str:
    try:
        compressed_bytes = bytes.fromhex(s)
        decompressed_bytes = zlib.decompress(compressed_bytes)
        return decompressed_bytes.decode("utf-8")
    except:  # noqa
        return s


def compressed_str_serializer(s: str) -> str:
    decompressed_bytes = s.encode("utf-8")
    return zlib.compress(decompressed_bytes).hex()


# this type will take a string and automatically compress and
# decompress it when it is serialized and deserialized
CompressedStr = Annotated[
    str,
    PlainValidator(compressed_str_validator),
    PlainSerializer(compressed_str_serializer),
    WithJsonSchema({"type": "string"}),
]


class MoleculeSpec(BaseModel):
    """A molecule schema to be output by OpenMMGenerators."""

    name: str
    count: int
    charge_scaling: float
    charge_method: str
    openff_mol: str  # a tk.Molecule object serialized with to_json

    # def as_dict(self) -> dict:
    #     return self.model_dump()

    # @classmethod
    # def from_dict(cls,dct):
    #     return cls(**dct)


@arrow_incompatible
class MDTaskDocument(BaseModel):  # type: ignore[call-arg]
    """Definition of the OpenMM task document."""

    tags: list[str] | None = Field(
        [], title="tag", description="Metadata tagged to a given task."
    )

    dir_name: str | None = Field(None, description="The directory for this MD task")

    state: TaskState | None = Field(None, description="State of this calculation")

    job_uuids: list[str] | None = Field(
        None,
        description="The job_uuids for all contributing jobs, this will only"
        "have a value if the taskdoc is generated by a Flow.",
    )

    calcs_reversed: list | None = Field(
        None,
        title="Calcs reversed data",
        description="Detailed data for each MD calculation contributing to "
        "the task document.",
    )

    interchange: CompressedStr | None = Field(
        None, description="An interchange object serialized to json."
    )

    mol_specs: list[MoleculeSpec] | None = Field(
        None,
        description="Molecules within the system. Only makes sense "
        "for molecular systems.",
    )

    structure: Structure | None = Field(
        None,
        title="Structure",
        description="The final structure for the simulation. Saved only "
        "if specified by job.",
    )

    force_field: str | None = Field(None, description="The classical MD forcefield.")

    task_type: str | None = Field(None, description="The type of calculation.")

    # task_label: str | None= Field(None, description="A description of the task")
    # TODO: where does task_label get added

    last_updated: datetime = Field(
        default_factory=utcnow,
        description="Timestamp for the most recent calculation for this task document",
    )

    model_config = ConfigDict(extra="allow")


class ClassicalMDTaskDocument(MDTaskDocument):
    """Definition of the OpenMM task document."""

    mol_specs: list[MoleculeSpec] | None = Field(
        None, description="Molecules within the system."
    )



import sys
import types
import typing
from collections.abc import Iterable, Mapping
from datetime import datetime
from enum import Enum
from pathlib import Path
from types import UnionType

import pyarrow as pa
import typing_extensions
from monty.json import MSONable
from pydantic._internal._model_construction import ModelMetaclass
from pydantic.types import ImportString

RED = "\033[31m"
BLUE = "\033[34m"
RESET = "\033[0m"

PY_PRIMITIVES_TO_ARROW = {
    int: pa.int64(),
    float: pa.float64(),
    str: pa.string(),
    bool: pa.bool_(),
    datetime: pa.timestamp("us"),
}


# issues re: pa.UnionType (sparse and dense) implementation:
# https://github.com/apache/arrow/issues/19157
# https://github.com/apache/arrow/issues/31211
# https://github.com/apache/arrow/issues/43857

# until union (de)serialization is supported on the pyarrow side we
# strictly must narrow all union types in emmet to a single type, e.g.,
#   my_list_of_numbers: list[int | float | None] = pydantic.Field(...)
# won't fly


def arrowize(obj) -> pa.DataType:
    """
    Converts Python type annotations to PyArrow data types.

    Recursively introspects type annotations and constructs corresponding
    PyArrow data types. This enables schema generation for Arrow-based
    serialization but does not guarantee successful serialization.

    Supported types:
    - Python primitives
    - Generic containers (with annotations)
    - Literals
    - Enum classes (with caveats)
    - Pydantic models
    - TypedDict
    - MSONable objects with type adapters

    Limitations:
    - Union types must resolve to a single primitive type
    - Container types must be homogeneous
    - Map keys cannot be Union types

    Args:
        obj: A type annotation, class, or object with type annotations

    Returns:
        A PyArrow DataType corresponding to the input type

    Note:
        Union type serialization is currently unsupported in PyArrow.
        All union types are narrowed to their first non-None member.
    """
    assert obj not in (
        list,
        tuple,
        set,
        dict,
        typing.Dict,
        typing.List,
        typing.Set,
        typing.Tuple,
    ), f"Cannot construct arrow type from container of type {RED}{obj}{RESET} without typed arguments"

    assert (
        obj is not typing.Any
    ), f"Cannot infer arrow type from: {RED}{obj}{RESET}, ambiguous ({BLUE}Any{RESET}) values cannot be resolved"

    if obj in PY_PRIMITIVES_TO_ARROW:
        return PY_PRIMITIVES_TO_ARROW[obj]

    if typing.get_origin(obj) in (list, tuple, set, Iterable):
        args = obj.__args__
        if len(args) == 1:
            return pa.list_(arrowize(args[0]))

        assert all(
            [
                member == args[0]
                for member in args
                if not isinstance(member, types.EllipsisType)
            ]
        ), f"Cannot infer arrow compatible primitive from iterable type containing mixed types: {RED}{obj}{RESET}"

        # once union type roundtripping is supported in pyarrow, something like this could work:
        #     return pa.dense_union(
        #         [
        #             pa.field(f"child_{idx}", arrowize(field))
        #             for idx, field, in zip(range(len(args)), args)
        #         ]
        #     )

        return pa.list_(arrowize(args[0]))

    if typing.get_origin(obj) in (Mapping, dict):
        args = typing.get_args(obj)

        assert all(
            not (
                isinstance(arg, typing._UnionGenericAlias | UnionType)  # type: ignore[attr-defined]
                and len(list(filter(lambda x: x is not types.NoneType, arg.__args__)))
                > 1
            )
            for arg in args
        ), f"""
        Cannot construct arrow map type from: {RED}{obj}{RESET}.
        Keys and values for map types must resolve to single primitive
        data type, not Union type:
            key annotation: {BLUE}{args[0]}{RESET}
            value annotation: {BLUE}{args[1]}{RESET}
        """

        return pa.map_(arrowize(args[0]), arrowize(args[1]))

    if typing.get_origin(obj) is typing.Literal:
        args = obj.__args__
        first_type = type(args[0])

        assert all(
            isinstance(member, first_type) for member in args
        ), f"Cannot infer arrow compatible primitive from literal with mixed types: {RED}{obj}{RESET}"

        return PY_PRIMITIVES_TO_ARROW[first_type]

    if typing.get_origin(obj) is typing.NotRequired:
        return arrowize(obj.__args__[0])

    if isinstance(obj, type) and issubclass(obj, Enum):
        # making an assumption here based on knowledge of emmet + pymatgen enum classes...
        # should have better way to handle all enums to get serialized representations
        return PY_PRIMITIVES_TO_ARROW[str]

    if isinstance(obj, typing._UnionGenericAlias | UnionType):  # type: ignore[attr-defined]
        arrow_types = [
            (arg.__name__, arrowize(arg))
            for arg in list(
                filter(
                    lambda x: x is not types.NoneType,  # noqa: E721
                    obj.__args__,
                )
            )
        ]

        assert all(
            [field == arrow_types[0][1] for (field_name, field) in arrow_types]
        ), f"""
        (De)Serialization of Union types is not supported in pyarrow currently,
        narrow the types of {RED}{obj}{RESET} to resolve to a single primitive
        """
        # once union type roundtripping is supported in pyarrow, something like this could work:
        #     return pa.dense_union(
        #         [pa.field(f"{field_name}", field) for (field_name, field) in arrow_types]
        #     )

        return arrow_types[0][1]

    if isinstance(obj, ModelMetaclass):
        return pa.struct(
            [
                pa.field(
                    field_name,
                    arrowize(
                        obj.type_overrides[field_name]
                        if hasattr(obj, "type_overrides")
                        and field_name in obj.type_overrides
                        else value.annotation
                    ),
                )
                for field_name, value in obj.model_fields.items()
                if not value.exclude
            ]
        )

    if any(obj is str_like for str_like in (ImportString, Path)):
        return PY_PRIMITIVES_TO_ARROW[str]

    if isinstance(obj, typing._TypedDictMeta | typing_extensions._TypedDictMeta):
        return pa.struct(
            [
                pa.field(field_name, arrowize(value))
                for field_name, value in obj.__annotations__.items()
            ]
        )

    if isinstance(obj, typing.ForwardRef):
        if sys.version_info >= (3, 12, 4):
            # ``type_params`` were added in 3.13 and the signature of _evaluate()
            # is not backward-compatible (it was backported to 3.12.4, so anything
            # before 3.12.4 still has the old signature).
            # See: https://github.com/python/cpython/pull/118104.
            return arrowize(
                obj._evaluate(globals(), locals(), {}, recursive_guard=frozenset())
            )
        return arrowize(obj._evaluate(globals(), locals(), frozenset()))

    if isinstance(obj, typing.TypeVar):
        return arrowize(obj.__constraints__[1])

    if isinstance(obj, typing._AnnotatedAlias):
        return arrowize(obj.__args__[0])

    if isinstance(obj, type) and issubclass(obj, MSONable):
        assert hasattr(
            obj, "__type_adapter__"
        ), f"""
        No serialization adapter is specified for msonable obj: {RED}{obj}{RESET}.
        TypeVars can be written for external classes in the types.pymatgen_types module.
        Internal classes may use the set_msonable_type_adapter decorator.
        """

        return arrowize(obj.__type_adapter__.model_fields["root"].annotation)

    if isinstance(obj, type):
        return arrowize(
            next(
                filter(
                    lambda primitive: issubclass(obj, primitive), PY_PRIMITIVES_TO_ARROW
                )
            )
        )



"""Utilities to determine level of theory, task type, and calculation type for Q-Chem calculations in the pydantic Docs paradigm"""

from emmet.core.qchem.calc_types import CalcType, LevelOfTheory, TaskType
from emmet.core.qchem.calc_types.calc_types import BASIS_SETS, FUNCTIONALS
from emmet.core.qchem.calculation import CalculationInput

__author__ = (
    "Evan Spotte-Smith <ewcspottesmith@lbl.gov>, Rishabh Debraj Guha <rdguha@lbl.gov>"
)

# TODO: better way of translating names
functional_synonyms = {
    "b97mv": "b97m-v",
    "b97mrv": "b97m-rv",
    "wb97xd": "wb97x-d",
    "wb97xd3": "wb97x-d3",
    "wb97xv": "wb97x-v",
    "wb97mv": "wb97m-v",
}

smd_synonyms = {
    "DIELECTRIC=7,230;N=1,410;ALPHA=0,000;BETA=0,859;GAMMA=36,830;PHI=0,000;PSI=0,000": "diglyme",
    "DIELECTRIC=18,500;N=1,415;ALPHA=0,000;BETA=0,735;GAMMA=20,200;PHI=0,000;PSI=0,000": "3:7 EC:EMC",
}


def level_of_theory(parameters: CalculationInput) -> LevelOfTheory:
    """

    Returns the level of theory for a calculation,
    based on the input parameters given to Q-Chem

    Args:
        parameters: dict of Q-Chem input parameters

    """

    funct_raw = parameters.rem.get("method")
    basis_raw = parameters.rem.get("basis")

    if funct_raw is None or basis_raw is None:
        raise ValueError(
            'Method and basis must be included in "rem" section ' "of parameters!"
        )

    disp_corr = parameters.rem.get("dft_d")

    if disp_corr is None:
        funct_lower = funct_raw.lower()
        funct_lower = functional_synonyms.get(funct_lower, funct_lower)
    else:
        # Replace Q-Chem terms for D3 tails with more common expressions
        disp_corr = disp_corr.replace("_bj", "(bj)").replace("_zero", "(0)")
        funct_lower = f"{funct_raw}-{disp_corr}"

    basis_lower = basis_raw.lower()

    # --> TODO: replace with enums
    functional = [f for f in FUNCTIONALS if f.lower() == funct_lower]
    if not functional:
        raise ValueError(f"Unexpected functional {funct_lower}!")

    functional = functional[0]

    basis = [b for b in BASIS_SETS if b.lower() == basis_lower]
    if not basis:
        raise ValueError(f"Unexpected basis set {basis_lower}!")
    # <--

    basis = basis[0]

    solvent_method = parameters.rem.get("solvent_method", "").lower()
    if solvent_method == "":
        solvation = "VACUUM"
    elif solvent_method in ["pcm", "cosmo"]:
        solvation = "PCM"
    # TODO: Add this once added into pymatgen and atomate
    # elif solvent_method == "isosvp":
    #     if parameters.get("svp", {}).get("idefesr", 0):
    #         solvation = "CMIRS"
    #     else:
    #         solvation = "ISOSVP"
    elif solvent_method == "smd":
        solvation = "SMD"
    else:
        raise ValueError(f"Unexpected implicit solvent method {solvent_method}!")

    lot = f"{functional}/{basis}/{solvation}"

    return LevelOfTheory(lot)


def solvent(parameters: CalculationInput, custom_smd: str | None = None) -> str:
    """
    Returns the solvent used for this calculation.

    Args:
        parameters: dict of Q-Chem input parameters
        custom_smd: (Optional) string representing SMD parameters for a
        non-standard solvent
    """

    lot = level_of_theory(parameters)
    solvation = lot.value.split("/")[-1]

    if solvation == "PCM":
        dielectric = float(parameters.get("solvent", {}).get("dielectric", 78.39))
        dielectric_string = f"{dielectric:.2f}".replace(".", ",")
        return f"DIELECTRIC={dielectric_string}"
    # TODO: Add this once added into pymatgen and atomate
    # elif solvation == "ISOSVP":
    #     dielectric = float(parameters.get("svp", {}).get("dielst", 78.39))
    #     rho = float(parameters.get("svp", {}).get("rhoiso", 0.001))
    #     return f"DIELECTRIC={round(dielectric, 2)},RHO={round(rho, 4)}"
    # elif solvation == "CMIRS":
    #     dielectric = float(parameters.get("svp", {}).get("dielst", 78.39))
    #     rho = float(parameters.get("svp", {}).get("rhoiso", 0.001))
    #     a = parameters.get("pcm_nonels", {}).get("a")
    #     b = parameters.get("pcm_nonels", {}).get("b")
    #     c = parameters.get("pcm_nonels", {}).get("c")
    #     d = parameters.get("pcm_nonels", {}).get("d")
    #     solvrho = parameters.get("pcm_nonels", {}).get("solvrho")
    #     gamma = parameters.get("pcm_nonels", {}).get("gamma")
    #
    #     string = f"DIELECTRIC={round(dielectric, 2)},RHO={round(rho, 4)}"
    #     for name, (piece, digits) in {"A": (a, 6), "B": (b, 6), "C": (c, 1), "D": (d, 3),
    #                                   "SOLVRHO": (solvrho, 2), "GAMMA": (gamma, 1)}.items():
    #         if piece is None:
    #             piecestring = "NONE"
    #         else:
    #             piecestring = f"{name}={round(float(piece), digits)}"
    #         string += "," + piecestring
    #     return string
    elif solvation == "SMD":
        solvent = parameters.get("smx", {}).get("solvent", "water")
        if solvent == "other":
            if custom_smd is None:
                raise ValueError(
                    "SMD calculation with solvent=other requires custom_smd!"
                )

            names = ["DIELECTRIC", "N", "ALPHA", "BETA", "GAMMA", "PHI", "PSI"]
            numbers = [float(x) for x in custom_smd.split(",")]

            string = ""
            for name, number in zip(names, numbers):
                string += f"{name}={number:.3f};"
            return string.rstrip(",").rstrip(";").replace(".", ",")
        else:
            return f"SOLVENT={solvent.upper()}"
    else:
        return "NONE"


def lot_solvent_string(
    parameters: CalculationInput, custom_smd: str | None = None
) -> str:
    """
    Returns a string representation of the level of theory and solvent used for this calculation.

    Args:
        parameters: dict of Q-Chem input parameters
        custom_smd: (Optional) string representing SMD parameters for a
        non-standard solvent
    """

    lot = level_of_theory(parameters).value
    solv = solvent(parameters, custom_smd=custom_smd)
    return f"{lot}({solv})"


def task_type(
    parameters: CalculationInput, special_run_type: str | None = None
) -> TaskType:
    if special_run_type == "frequency_flattener":
        return TaskType("Frequency Flattening Geometry Optimization")
    elif special_run_type == "ts_frequency_flattener":
        return TaskType("Frequency Flattening Transition State Geometry Optimization")

    if parameters.job_type == "sp":
        return TaskType("Single Point")
    elif parameters.job_type == "force":
        return TaskType("Force")
    elif parameters.job_type == "opt":
        return TaskType("Geometry Optimization")
    elif parameters.job_type == "ts":
        return TaskType("Transition State Geometry Optimization")
    elif parameters.job_type == "freq":
        return TaskType("Frequency Analysis")

    return TaskType("Unknown")


def calc_type(
    parameters: CalculationInput, special_run_type: str | None = None
) -> CalcType:
    """
    Determines the calc type

    Args:
        inputs: inputs dict with an incar, kpoints, potcar, and poscar dictionaries
        parameters: Dictionary of VASP parameters from Vasprun.xml
    """
    rt = level_of_theory(parameters).value
    tt = task_type(parameters, special_run_type=special_run_type).value
    return CalcType(f"{rt} {tt}")



from collections.abc import Callable
from enum import Enum, auto
from typing import Annotated, Any, TypeVar, ValuesView

import orjson
from pydantic import BeforeValidator, WrapSerializer
from pymatgen.analysis.phase_diagram import PhaseDiagram
from typing_extensions import TypedDict

from emmet.core.types.pymatgen_types.computed_entries_adapter import (
    TypedComputedStructureEntryDict,
)


class Mode(Enum):
    SHRED = auto()
    STITCH = auto()


TypedMSONableElementDict = TypedDict(
    "TypedMSONableElementDict", {"@module": str, "@class": str, "element": str}
)

TypedSimplexDict = TypedDict(
    "TypedSimplexDict",
    {"@module": str, "@class": str, "@version": str, "coords": list[list[float]]},
)

TypedComputedDataDict = TypedDict(
    "TypedComputedDataDict",
    {
        "dim": int,
        # "el_refs": list[[str, ComputedStructureEntry]],
        "el_refs_elements": list[str],
        "el_refs_entries": list[TypedComputedStructureEntryDict],
        "facets": list[list[int]],
        "qhull_data": list[list[float]],
        "qhull_entries": list[TypedComputedStructureEntryDict],
        "simplexes": list[TypedSimplexDict],
        "all_entries": list[TypedComputedStructureEntryDict],
    },
)

TypedPhaseDiagramDict = TypedDict(
    "TypedPhaseDiagramDict",
    {
        "@module": str,
        "@class": str,
        "all_entries": list[TypedComputedStructureEntryDict],
        "elements": list[TypedMSONableElementDict],
        "computed_data": TypedComputedDataDict,
    },
)

PhaseDiagramTypeVar = TypeVar(
    "PhaseDiagramTypeVar", PhaseDiagram, TypedPhaseDiagramDict
)


def entries_list_serde(entries_list: ValuesView | list[dict], serde_fn: Callable):
    for entry in entries_list:
        entry["energy_adjustments"] = serde_fn(entry["energy_adjustments"])


def phase_diagram_serde(d: dict, mode: Mode, serde_fn: Callable):
    entries_list_serde(d["all_entries"], serde_fn)
    for key in ["all_entries", "qhull_entries"]:
        entries_list_serde(d["computed_data"][key], serde_fn)

    match mode:
        case Mode.SHRED:
            el_ref_pairs = d["computed_data"].pop("el_refs")

            elements = []
            el_refs_entries = []
            for element, entry in el_ref_pairs:
                elements.append(str(element))
                el_refs_entries.append(entry.as_dict())

            entries_list_serde(el_refs_entries, serde_fn)

            d["computed_data"]["el_refs_elements"] = elements
            d["computed_data"]["el_refs_entries"] = el_refs_entries

        case Mode.STITCH:
            elements = d["computed_data"].pop("el_refs_elements")
            el_refs_entries = d["computed_data"].pop("el_refs_entries")

            entries_list_serde(el_refs_entries, serde_fn)

            d["computed_data"]["el_refs"] = [
                (i, j) for i, j in zip(elements, el_refs_entries)
            ]


def phase_diagram_serializer(phase_diagram, nxt, info) -> dict[str, Any]:
    default_serialized_object = nxt(phase_diagram.as_dict(), info)

    for key in ["all_entries", "qhull_entries", "simplexes"]:
        default_serialized_object["computed_data"][key] = [
            entry.as_dict() for entry in default_serialized_object["computed_data"][key]
        ]

    for simplex in default_serialized_object["computed_data"]["simplexes"]:
        simplex["coords"] = simplex["coords"].tolist()

    format = info.context.get("format") if info.context else None
    if format == "arrow":
        phase_diagram_serde(
            default_serialized_object, mode=Mode.SHRED, serde_fn=orjson.dumps
        )

    return default_serialized_object


def phase_diagram_deserializer(value) -> PhaseDiagram:
    if isinstance(value, dict):
        if all(
            key in value["computed_data"]
            for key in ["el_refs_elements", "el_refs_entries"]
        ):
            phase_diagram_serde(value, mode=Mode.STITCH, serde_fn=orjson.loads)
        return PhaseDiagram.from_dict(value)
    return value


PhaseDiagramType = Annotated[
    PhaseDiagramTypeVar,
    BeforeValidator(phase_diagram_deserializer),
    WrapSerializer(phase_diagram_serializer),
]



import pytest
from pymatgen.core import Structure

from . import test_structures
from emmet.core import ARROW_COMPATIBLE
from emmet.core.chemenv import ChemEnvDoc

if ARROW_COMPATIBLE:
    import pyarrow as pa

    from emmet.core.arrow import arrowize


@pytest.mark.parametrize("structure", test_structures.values())
def test_chemenv(structure: Structure):
    """Very simple test to make sure this actually works"""
    print(f"Should work : {structure.composition}")
    doc = ChemEnvDoc.from_structure(
        structure=structure, material_id=33, deprecated=False
    )
    valences = [getattr(site.specie, "oxi_state", None) for site in structure]
    valences = [v for v in valences if v is not None]
    if len(valences) == len(structure):
        assert doc.model_dump()["warnings"] is None
    # elif structure.composition.almost_equals(Composition("CsCl")):
    #    # We do not have reference polyhedra above a certain number of neighbors.
    #    # ChemEnv cannot deliver an answer without oxidation states.
    #    assert doc.model_dump()["warnings"] == "ChemEnv algorithm failed."
    else:
        assert (
            doc.model_dump()["warnings"]
            == "No oxidation states available. Cation-anion bonds cannot be identified."
        )


@pytest.mark.skipif(
    not ARROW_COMPATIBLE, reason="pyarrow must be installed to run this test."
)
def test_arrow(
    structure=next(iter(test_structures.values())),
):
    doc = ChemEnvDoc.from_structure(
        structure=structure, material_id=33, deprecated=False
    )

    arrow_struct = pa.scalar(
        doc.model_dump(context={"format": "arrow"}), type=arrowize(ChemEnvDoc)
    )
    test_arrow_doc = ChemEnvDoc(**arrow_struct.as_py(maps_as_pydicts="strict"))

    assert doc == test_arrow_doc



import pytest
from emmet.core.oxidation_states import OxidationStateDoc, OxiStateAssigner
from pymatgen.core import Structure

from . import test_structures


@pytest.mark.parametrize("structure", test_structures.values())
def test_oxidation_state(structure: Structure):
    """Very simple test to make sure this actually works"""
    struct_has_charges_assigned = (
        structure.composition.copy().remove_charges() != structure.composition
    )

    for method in (
        None,
        "bva",
    ):

        doc = OxidationStateDoc.from_structure(
            structure, material_id=33, deprecated=False
        )
        assert doc is not None
        assert doc.structure is not None

        if struct_has_charges_assigned and not method:
            assert doc.method == OxiStateAssigner.MANUAL
        elif doc.possible_valences:
            assert doc.method in {OxiStateAssigner.BVA, OxiStateAssigner.GUESS}
        else:
            assert doc.method is None



"""Task types and level of theory components for Q-Chem calculations"""

from importlib.resources import files as import_resource_files
from monty.serialization import loadfn

__author__ = "Evan Spotte-Smith <ewcspottesmith@lbl.gov>"

_calc_type_config = loadfn(
    str(import_resource_files("emmet.core.qchem.calc_types") / "calc_types.yaml")
)

# NB: this would be easier with setattr but the following is less opaque
FUNCTIONAL_CLASSES = _calc_type_config.get("FUNCTIONAL_CLASSES")
TASK_TYPES = _calc_type_config.get("TASK_TYPES")
BASIS_SETS = _calc_type_config.get("BASIS_SETS")
SOLVENT_MODELS = _calc_type_config.get("SOLVENT_MODELS")

FUNCTIONALS = [rt for functionals in FUNCTIONAL_CLASSES.values() for rt in functionals]



Vector3D = tuple[float, float, float]

Matrix3D = tuple[Vector3D, Vector3D, Vector3D]

Vector6D = tuple[float, float, float, float, float, float]

MatrixVoigt = tuple[Vector6D, Vector6D, Vector6D, Vector6D, Vector6D, Vector6D]

Tensor3R = list[list[list[float]]]

Tensor4R = list[list[list[list[float]]]]

ListVector3D = list[float]

ListMatrix3D = list[ListVector3D]

VOIGT_INDICES = [(0, 0), (1, 1), (2, 2), (1, 2), (0, 2), (0, 1)]


def matrix_3x3_to_voigt(matrix: Matrix3D) -> list[float]:
    """Convert a 3x3 symmetric matrix to its Voigt vector representation.

    Parameters
    -----------
    matrix : Matrix3D

    Returns
    -----------
    list of float
    """
    return [matrix[idx[0]][idx[1]] for idx in VOIGT_INDICES]
